from torchvision.transforms import Compose, RandomHorizontalFlip, RandomCrop, RandomRotation, ToTensor, CenterCrop, Normalize
from typing import Any

import torch
from torchvision import transforms
from torchvision.datasets import ImageNet

from avalanche.benchmarks.datasets import default_dataset_location
from avalanche.benchmarks import dacCLScenario, SplitCIFAR100
from avalanche.benchmarks import CLExperience, CLStream
from avalanche.benchmarks.utils.dataset_utils import ConstantSequence

from avalanche.benchmarks.utils.avalanche_dataset import _TaskSubsetDict, AvalancheConcatDataset
from avalanche.benchmarks import CORe50, nc_benchmark
from avalanche.benchmarks.classic.ccifar100 import _get_cifar100_dataset, _default_cifar100_train_transform, _default_cifar100_eval_transform
from .single_image_data import SingleImageDataset, FastSingleImageDataset

SEED_BENCHMARK_RUNS = [1234, 2345, 3456, 5678, 6789]  # 5 different seeds to randomize class orders

fast_data_dir = "$ANON_HOME_DIR/DATA"


def seq_cifar100_benchmark(num_experiences, run_id):
    cifar_train, cifar_test = _get_cifar100_dataset(fast_data_dir + "/cifar100")
    return nc_benchmark(
        train_dataset=cifar_train,
        test_dataset=cifar_test,
        n_experiences=num_experiences,
        task_labels=True,
        seed=SEED_BENCHMARK_RUNS[run_id],
        shuffle=True,
        class_ids_from_zero_in_each_exp=False,
        train_transform=_default_cifar100_train_transform,
        eval_transform=_default_cifar100_eval_transform,
    )


def seq_core50_benchmark(scenario, run_id):
    """Seq CL scenario on CoRE50."""
    assert scenario in {
        "ni",
        "joint",
        "nc",
    }, "`scenario` argument must be one of {'ni', 'joint', 'nc'}."

    core50_normalization = Normalize(
        mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]
    )
    core50_train_transforms = Compose(
        [
            RandomHorizontalFlip(p=0.5),
            RandomCrop(size=128, padding=1),
            RandomRotation(15),
            ToTensor(),
            core50_normalization,
        ]
    )
    core50_eval_transforms = Compose(
        [CenterCrop(size=128), ToTensor(), core50_normalization]
    )

    if scenario == "ni":
        benchmark = CORe50(
            scenario="ni",
            train_transform=core50_train_transforms,
            eval_transform=core50_eval_transforms,
            run=run_id,
            dataset_root=fast_data_dir + "/core50",
        )
    elif scenario == "nc":
        benchmark = CORe50(
            scenario="nc",
            train_transform=core50_train_transforms,
            eval_transform=core50_eval_transforms,
            run=run_id,
            dataset_root=fast_data_dir + "/core50",
        )
    elif scenario == "joint":
        core50nc = CORe50(scenario="nc", dataset_root=fast_data_dir + "/core50")
        train_cat = AvalancheConcatDataset(
            [e.dataset for e in core50nc.train_stream]
        )
        test_cat = AvalancheConcatDataset(
            [e.dataset for e in core50nc.test_stream]
        )
        benchmark = nc_benchmark(
            train_cat, test_cat, n_experiences=1, task_labels=True
        )
    else:
        assert False, "Should never get here."

    # add task labels
    benchmark.train_stream = as_multitask_stream(benchmark.train_stream)
    benchmark.test_stream = as_multitask_stream(benchmark.test_stream)
    return benchmark


class UnlabeledImageNet(ImageNet):
    def __init__(self, img_size):
        """ Used as auxiliary data for MACL CL with COrE50 (128x128 images). """
        root = default_dataset_location('imagenet')
        self.img_size = img_size

        transform = transforms.Compose([
            transforms.CenterCrop(224),
            transforms.Resize(img_size),
            transforms.ToTensor()
        ])
        super().__init__(root, split="train", transform=transform)

    def __getitem__(self, index):
        return super().__getitem__(index)[0]


class ImageNet32(ImageNet):
    def __init__(self, root: str = None, split: str = 'train', **kwargs: Any):
        normalize = transforms.Normalize(mean=[0.485, 0.456, 0.406],
                                         std=[0.229, 0.224, 0.225])

        if root is None:
            root = default_dataset_location('imagenet')

        if split == 'train':
            transform = transforms.Compose([
                transforms.RandomResizedCrop(224),
                transforms.Resize(32),
                transforms.RandomHorizontalFlip(),
                transforms.ToTensor(),
                normalize,
            ])
        else:
            transform = transforms.Compose([
                transforms.Scale(256),
                transforms.CenterCrop(224),
                transforms.Resize(32),
                transforms.ToTensor(),
                normalize,
            ])
        super().__init__(root, split, transform=transform, **kwargs)


class ImageNet128(ImageNet):
    def __init__(self, root: str = None, split: str = 'train', **kwargs: Any):
        """ Used as auxiliary data for MACL CL with COrE50 (128x128 images). """
        normalize = transforms.Normalize(mean=[0.485, 0.456, 0.406],
                                         std=[0.229, 0.224, 0.225])

        if root is None:
            root = default_dataset_location('imagenet')

        if split == 'train':
            transform = transforms.Compose([
                transforms.RandomSizedCrop(224),
                transforms.Resize(128),
                transforms.RandomHorizontalFlip(),
                transforms.ToTensor(),
                normalize,
            ])
        else:
            transform = transforms.Compose([
                transforms.Scale(256),
                transforms.CenterCrop(224),
                transforms.Resize(128),
                transforms.ToTensor(),
                normalize,
            ])
        super().__init__(root, split, transform=transform, **kwargs)


class maclCIFAR100(dacCLScenario):
    """macl scenario on CIFAR100.

    The pretrained models and class splits are taken from
    https://arxiv.org/abs/2112.06511
    """

    def __init__(self, scenario="split", run_id=0):
        """Init.

        :param scenario: If 'split', use a class-incremental scenario with 5
            experiences (2 classes each). If 'joint', use a single experience
            with all the classes. This should be used only as a baseline since
            it is not a continual scenario.
        :param run_id: an integer in [0, 4]. Each run uses a different set of
            expert models and data splits.
        """
        # assert run_id == 0, "single run in Split CIFAR100"

        assert scenario in {
            "split",
            "joint",
        }, "`scenario` argument must be one of {'split', 'joint'}."

        CURR_SEED = SEED_BENCHMARK_RUNS[run_id]

        if scenario == "split":
            benchmark = SplitCIFAR100(
                n_experiences=10, return_task_id=False, seed=CURR_SEED
            )
        elif scenario == "joint":
            benchmark = SplitCIFAR100(
                n_experiences=1, return_task_id=False, seed=CURR_SEED
            )
        else:
            assert False, "Should never get here."

        len_stream = len(benchmark.train_stream)
        scenario_name = f"{scenario}_cifar100"
        basedir = f"$LOGDIR/logs/resnet/{scenario_name}"

        experts_stream = []
        for i in range(len_stream):
            fname_i = f"{basedir}/model_e{i}.pt"
            model = torch.load(fname_i)
            model.to("cpu").eval()
            experts_stream.append(model)
        super().__init__(benchmark, experts_stream)


if __name__ == '__main__':
    from avalanche.training import Naive

    bm = maclCIFAR100()
    for ii, exp in enumerate(bm.expert_models_stream):
        # print(exp.expert_model)
        strat = Naive(exp.expert_model, None, eval_mb_size=512)
        strat.eval(bm.original_benchmark.test_stream[ii], num_workers=8)


def as_multitask_benchmark(benchmark):
    """Adds task label info to experiences in a benchmark.
    in-place operation.
    """
    for s in benchmark.streams.values():
        for ii, exp in enumerate(s):
            exp.task_label = ii
            exp.task_labels = [ii]
    return benchmark


def as_multitask_stream(ss):
    """Adds task labels to a single task supervised stream."""
    new_exps = []
    for ii, e in enumerate(ss):
        ne = CLExperience()
        ne.classes_in_this_experience = e.classes_in_this_experience
        ne.task_labels = [ii]
        ne.task_label = ii
        ne.dataset = e.dataset

        ne.dataset.targets_task_labels = ConstantSequence(ii, len(ne.dataset))

        told_idx = ne.dataset.tasks_pattern_indices
        del ne.dataset.tasks_pattern_indices[0]
        ne.dataset.tasks_pattern_indices[ii] = told_idx

        task_dict = _TaskSubsetDict()
        task_idxs = ne.dataset.tasks_pattern_indices[ii]
        task_dict[ii] = (ne.dataset, task_idxs)
        ne.dataset.val_to_idx = task_dict

        new_exps.append(ne)
    return CLStream(f"original_{ss.name}", new_exps)
