from PIL import Image
from torchvision.utils import save_image
import torchvision.transforms as tfs

import torch
import math

from PIL import ImageFilter
import random
import kornia.augmentation as kaug


class GaussianBlur(object):
    """Gaussian blur augmentation from SimCLR https://arxiv.org/abs/2002.05709"""

    def __init__(self, sigma=[.1, 2.]):
        self.sigma = sigma

    def __call__(self, x):
        sigma = random.uniform(self.sigma[0], self.sigma[1])
        x = x.filter(ImageFilter.GaussianBlur(radius=sigma))
        return x


class MyRandomResizedCrop(tfs.RandomResizedCrop):
    @staticmethod
    def get_params(
            img, scale, ratio):
        """Get parameters for ``crop`` for a random sized crop.

        # compared to original: there's no fallback of doing a CentralCrop

        Args:
            img (PIL Image or Tensor): Input image.
            scale (list): range of scale of the origin size cropped
            ratio (list): range of aspect ratio of the origin aspect ratio cropped

        Returns:
            tuple: params (i, j, h, w) to be passed to ``crop`` for a random
                sized crop.
        """
        width, height = img.size[0], img.size[1]
        area = height * width

        while True:  # basically no falling back to center-crop, as in the original implementation
            target_area = area * (
                torch.empty(1).uniform_(scale[0], scale[1]).item()  # **2 # the cubed is new
            )
            log_ratio = torch.log(torch.tensor(ratio))
            aspect_ratio = torch.exp(
                torch.empty(1).uniform_(log_ratio[0], log_ratio[1])
            ).item()

            w = int(round(math.sqrt(target_area * aspect_ratio)))
            h = int(round(math.sqrt(target_area / aspect_ratio)))

            if 0 < w <= width and 0 < h <= height:
                i = torch.randint(0, height - h + 1, size=(1,)).item()
                j = torch.randint(0, width - w + 1, size=(1,)).item()
                return i, j, h, w


class SingleImageDataset(torch.utils.data.Dataset):
    def __init__(self, image_path, initcrop=0.5, degrees=30, scale=(0.01, 1.0), shear=30, vflip=False,
                 crop_size=256, mean=(0, 0, 0), std=(1, 1, 1), randinterp=False, cropfirst=True, debug=False,
                 length=int(1.281e6)):
        """Dataset class that combines the following transformations to generate samples from A SINGLE IMAGE

        1. RandomCrop (initialcrop) if cropfirst
        2. Random-Resized-Crop  if scale[0] > 0
            * size=(sqrt(2)*cropsize)            # sqrt(2) in the case of 45deg rotation.
            * scale=(1/scale[0], 1/scale[1])
        3. RandomAffine (degrees and shear) if shear!=0 and degrees!=0
        4. VerticalFlip (p=0.5) if vflip
        5. RandomHorizontalflip (p=0.5)
        6. CenterCrop (cropsize)
        7. ColorJitter
        #  8. RandomGaussianBlur # from MoCo-v2 repo
        9. ...
        """
        self.img = None
        self.image_path = image_path
        normalize = tfs.Normalize(mean=mean, std=std)
        self.img = (Image.open(self.image_path).convert('RGB'))

        scale = (1. / scale[0], 1. / scale[1])
        if randinterp and not debug:
            resizedcrop = tfs.RandomApply(
                [MyRandomResizedCrop(int(math.sqrt(2) * crop_size), scale=scale, interpolation=1),
                 MyRandomResizedCrop(int(math.sqrt(2) * crop_size), scale=scale, interpolation=2),
                 MyRandomResizedCrop(int(math.sqrt(2) * crop_size), scale=scale, interpolation=3)])
        else:
            resizedcrop = MyRandomResizedCrop(int(math.sqrt(2) * crop_size), scale=scale, interpolation=3)
        self.length = length  # run out after one IN-1k epoch
        self.crop_size = crop_size
        self.totensor = tfs.ToTensor()

        tfslist = []
        if cropfirst:
            tfslist.append(tfs.RandomCrop(int(initcrop * min(self.img.size))))

        if scale[0] > 0:
            tfslist.append(resizedcrop)
        if shear != 0:
            tfslist.append(tfs.RandomAffine(degrees, translate=None, scale=None, shear=shear, resample=Image.BILINEAR,
                                            fillcolor=0))
        elif degrees != 0:
            tfslist.append(tfs.RandomRotation(degrees=degrees, resample=Image.BILINEAR))

        if vflip:
            tfslist.append(tfs.RandomVerticalFlip(p=0.5))

        tfslist.append(tfs.CenterCrop(crop_size))
        if debug:
            tfslist.append(tfs.Resize((32, 32)))
        else:
            tfslist.extend([
                tfs.RandomApply([tfs.ColorJitter(brightness=0.4, contrast=0.4, saturation=0.4, hue=0.1)], 0.5),
                # GaussianBlur()
            ])
        tfslist.append(tfs.ToTensor())
        if mean != [0, 0, 0]:
            tfslist.append(normalize)
        self.transforms = tfs.Compose(tfslist)
        print("transforms:", tfslist)

    def __len__(self):
        return self.length

    def __getitem__(self, idx):
        if self.img is None:
            self.img = Image.open(self.image_path).convert('RGB')
        return self.transforms(self.img)


class FastSingleImageDataset(torch.utils.data.Dataset):
    def __init__(self, image_path, initcrop=0.5, degrees=30, scale=(0.01, 1.0), shear=30, vflip=False,
                 crop_size=256, mean=(0, 0, 0), std=(1, 1, 1), randinterp=False, cropfirst=True, debug=False,
                 length=int(1.281e6)):
        """WARNING: DON'T USE THIS. KORNIA IS SLOWER.
        See experiments/profile_dl.py for profiling results.

        Dataset class that combines the following transformations to generate samples from A SINGLE IMAGE

        1. RandomCrop (initialcrop) if cropfirst
        2. Random-Resized-Crop  if scale[0] > 0
            * size=(sqrt(2)*cropsize)            # sqrt(2) in the case of 45deg rotation.
            * scale=(1/scale[0], 1/scale[1])
        3. RandomAffine (degrees and shear) if shear!=0 and degrees!=0
        4. VerticalFlip (p=0.5) if vflip
        5. RandomHorizontalflip (p=0.5)
        6. CenterCrop (cropsize)
        7. ColorJitter
        #  8. RandomGaussianBlur # from MoCo-v2 repo
        9. ...
        """
        self.img = None
        self.image_path = image_path
        self.totensor = tfs.ToTensor()
        self.img = self.totensor(Image.open(self.image_path).convert('RGB'))# .to('cuda')

        self.length = length  # run out after one IN-1k epoch
        self.crop_size = crop_size

        normalize = tfs.Normalize(mean=mean, std=std)
        scale = (1. / scale[0], 1. / scale[1])
        x = int(math.sqrt(2) * crop_size)
        if randinterp and not debug:
            resizedcrop = tfs.RandomApply(
                [kaug.RandomResizedCrop((x, x), scale=scale, resample="bilinear"),
                 kaug.RandomResizedCrop((x, x), scale=scale, resample="bilinear"),
                 kaug.RandomResizedCrop((x, x), scale=scale, resample="bilinear")])
        else:
            resizedcrop = kaug.RandomResizedCrop((x, x), scale=scale, resample="bilinear")

        tfslist = []
        if cropfirst:
            x = int(initcrop * min(self.img.shape[1:]))
            tfslist.append(kaug.RandomCrop((x, x)))

        if scale[0] > 0:
            tfslist.append(resizedcrop)
        if shear != 0:
            tfslist.append(kaug.RandomAffine(degrees, translate=None, scale=None, shear=shear, resample="bilinear"))
        elif degrees != 0:
            tfslist.append(kaug.RandomRotation(degrees=degrees, resample=Image.BILINEAR))

        if vflip:
            tfslist.append(kaug.RandomVerticalFlip(p=0.5))

        tfslist.append(kaug.CenterCrop(crop_size))
        if debug:
            tfslist.append(kaug.Resize((32, 32)))
        else:
            tfslist.extend([tfs.RandomApply([kaug.ColorJitter(
                brightness=0.4, contrast=0.4, saturation=0.4, hue=0.1)], 0.5)])
        if mean != [0, 0, 0]:
            tfslist.append(normalize)
        self.transforms = tfs.Compose(tfslist)
        print("transforms:", tfslist)

    def __len__(self):
        return self.length

    def __getitem__(self, idx):
        aug_img = self.img.unsqueeze(0)
        aug_img = self.transforms(aug_img)
        return aug_img.squeeze(0)

    def get_batch(self, batch_size):
        aug_img = self.img.unsqueeze(0).repeat(batch_size, 1, 1, 1)
        aug_img = self.transforms(aug_img)
        return aug_img


def get_mono_data_loader(batch_size, num_workers, **kwargs):
    m = SingleImageDataset(**kwargs)
    return torch.utils.data.DataLoader(m, batch_size=batch_size, num_workers=num_workers,
                                       drop_last=True)


def save_batch_imgs(imgs, count, pth):
    for k in range(imgs.size(0)):
        save_image(imgs[k, :, :, :], pth + '/patch_%s.jpeg' % (k + count), padding=0, normalize=False, scale_each=False)

