import numpy as np
import torch


def rand_bbox(size, lam):
    """Sample random bounding box."""
    # To sample the binary mask M, we first sample the
    # bounding box coordinates B = (rx, ry , rw, rh) indicating
    # the cropping regions on xA and xB . The region B in xA is
    # removed and filled in with the patch cropped from B of xB .
    W = size[2]
    H = size[3]
    cut_rat = np.sqrt(1. - lam)
    cut_w = np.int(W * cut_rat)
    cut_h = np.int(H * cut_rat)

    # In our experiments, we sample rectangular masks M
    # whose aspect ratio is proportional to the original image. The
    # box coordinates are uniformly sampled according to:
    # uniform
    cx = np.random.randint(W)
    cy = np.random.randint(H)

    bbx1 = np.clip(cx - cut_w // 2, 0, W)
    bby1 = np.clip(cy - cut_h // 2, 0, H)
    bbx2 = np.clip(cx + cut_w // 2, 0, W)
    bby2 = np.clip(cy + cut_h // 2, 0, H)

    return bbx1, bby1, bbx2, bby2


class CutMix:
    def __init__(self, beta=1.0):
        """CutMix augmentation.

        The goal of CutMix is to generate a
        new training sample ( ̃x,  ̃y) by combining two training sam-
        ples (xA, yA) and (xB , yB )

        This implementation IGNORES CLASS LABELS, because I don't need them for MACL.
        Original pytorch implementation from the authors @ https://github.com/clovaai/CutMix-PyTorch/blob/master/train.py

        :param beta:
        """
        self.beta = beta

    @torch.no_grad()
    def __call__(self, x, target):
        """CutMix augmentation.

        :param x: batch of images <B, C, W, H>
        :param target: targets <B, K>
        :return:
        """
        assert len(x.shape) == 4
        # generate mixed sample
        # the combination ratio λ between two data
        # points is sampled from the beta distribution Beta(α, α)
        # In all our (original paper) experiments, we set α to 1, that is λ is sampled from
        # the uniform distribution (0, 1)
        lam = np.random.beta(self.beta, self.beta)

        # shuffle mini-batch to obtain pair of examples
        rand_index = torch.randperm(x.shape[0]).cuda()
        x_a, x_b = x, x[rand_index]
        target_a, target_b = target, target[rand_index]

        bbx1, bby1, bbx2, bby2 = rand_bbox(x.size(), lam)
        # The region B in xA is removed and filled in with the patch cropped from B of xB .
        x[:, :, bbx1:bbx2, bby1:bby2] = x_b[:, :, bbx1:bbx2, bby1:bby2]
        # adjust lambda to exactly match pixel ratio
        lam = 1 - ((bbx2 - bbx1) * (bby2 - bby1) / (x.shape[-1] * x.shape[-2]))
        # compute output
        return x, target_a, target_b, lam


if __name__ == '__main__':
    from avalanche.benchmarks import CORe50
    import matplotlib.pyplot as plt

    print("Just plotting a couple of images to check the augmentation is working.")
    data = CORe50().train_stream[0].dataset
    cm = CutMix()

    mb = data[:10]
    x, y_a, y_b, lam = cm(mb[0], mb[1])
    for i in range(10):
        plt.imshow(x.detach().cpu().numpy())
