from rich import print
import sys
from typing import List, TYPE_CHECKING

import torch
from torchvision.utils import save_image

from avalanche.evaluation.metric_results import MetricValue, TensorImage
from avalanche.logging import BaseLogger
from avalanche.evaluation.metric_utils import stream_type

from rich.table import Table
from rich.live import Live

from tqdm import tqdm

if TYPE_CHECKING:
    from avalanche.training import BaseStrategy


class RichTextLogger(BaseLogger):
    """
    The `TextLogger` class provides logging facilities
    printed to a user specified file. The logger writes
    metric results after each training epoch, evaluation
    experience and at the end of the entire evaluation stream.

    .. note::
        To avoid an excessive amount of printed lines,
        this logger will **not** print results after
        each iteration. If the user is monitoring
        metrics which emit results after each minibatch
        (e.g., `MinibatchAccuracy`), only the last recorded
        value of such metrics will be reported at the end
        of the epoch.

    .. note::
        Since this logger works on the standard output,
        metrics producing images or more complex visualizations
        will be converted to a textual format suitable for
        console printing. You may want to add more loggers
        to your `EvaluationPlugin` to better support
        different formats.
    """
    def __init__(self, logdir, file=sys.stdout, interactive=True):
        """
        Creates an instance of `TextLogger` class.

        :param logdir: a directory to save metric value not serializable to
            strings (e.g. images).
        :param file: destination file to which print metrics
            (default=sys.stdout).
        :param interactive: whether to use the interactive display mode,
            suggested for printing in the terminal. Disable it if you want to
            log the output in a logging file to reduce the visual clutter such
            as progress bars. (default=True)
        """
        super().__init__()
        self.logdir = logdir
        self.file = file
        self.interactive = interactive

        # state
        self.metric_vals = {}
        self.live = None
        self.curr_table = None
        self.is_table_initialized = False
        self._pbar = None

    def log_single_metric(self, name, value, x_plot):
        self.metric_vals[name] = (name, x_plot, value)

    def _val_to_str(self, m_val):
        if isinstance(m_val, torch.Tensor):
            return '\n' + str(m_val)
        elif isinstance(m_val, float):
            return f'{m_val:.4f}'
        else:
            return str(m_val)

    def print_current_metrics(self, strategy, metric_values):
        if strategy.is_training:
            self.train_print_current_metrics(metric_values)
        else:
            self.eval_print_current_metrics(strategy, metric_values)

    def train_print_current_metrics(self, metric_values):
        sorted_vals = sorted(metric_values, key=lambda x: x.name)
        for mval in sorted_vals:
            name, x, val = mval.name, mval.x_plot, mval.value
            if isinstance(val, TensorImage):
                str_name = '_'.join(name.split('/'))
                fname = f'{self.logdir}/{str_name}_x_{x}.png'
                save_image(val.image, fname)
                val = f'<serialized to {fname}>.'
            else:
                val = self._val_to_str(val)
            name = name.split('/')[0]
            print(f'\t{name} = {val}', file=self.file, flush=True)

    def eval_print_current_metrics(self, strategy, metric_values):
        if not self.is_table_initialized:
            sorted_vals = sorted(metric_values, key=lambda x: x.name)
            self.curr_table.add_column("Exp/Task")
            for mval in sorted_vals:
                name, x, val = mval.name, mval.x_plot, mval.value
                # keep only the metric name.
                # Other info is redundant
                name = name.split('/')[0]
                self.curr_table.add_column(name)
            self.is_table_initialized = True

        vals = {}
        exp_id = strategy.experience.current_experience
        task_id = strategy.experience.task_label
        vals["Exp/Task"] = f"E{exp_id}T{task_id}"
        for mval in metric_values:
            name, x, val = mval.name, mval.x_plot, mval.value
            name = name.split('/')[0]
            if isinstance(val, TensorImage):
                str_name = '_'.join(name.split('/'))
                fname = f'{self.logdir}/{str_name}_x_{x}.png'
                save_image(val.image, fname)
                val = f'<serialized to {fname}>.'
            else:
                val = self._val_to_str(val)
            vals[name] = val

        rowvals = []
        for col in self.curr_table.columns:
            if col.header in vals:
                rowvals.append(vals[col.header])
            else:
                rowvals.append(' - ')

        self.curr_table.add_row(*rowvals)
        self.live.refresh()
        # print(self.curr_table)

    def after_training_epoch(self, strategy: 'BaseStrategy',
                             metric_values: List['MetricValue'], **kwargs):
        self._end_progress()
        print(f'\t[italic]Epoch {strategy.clock.train_exp_epochs} ended.', file=self.file, flush=True)
        self.print_current_metrics(strategy, metric_values)
        self.metric_vals = {}

    def after_eval_exp(self, strategy: 'BaseStrategy',
                       metric_values: List['MetricValue'], **kwargs):
        self.print_current_metrics(strategy, metric_values)
        self.metric_vals = {}

    def before_training(self, strategy: 'BaseStrategy',
                        metric_values: List['MetricValue'], **kwargs):
        print('[bold blue]Start of training phase', file=self.file, flush=True)

    def before_eval_exp(self, strategy: 'BaseStrategy',
                    metric_values: List['MetricValue'], **kwargs):
        if self.live is None:
            print('[bold blue]Start of eval phase', file=self.file, flush=True)

            stream = stream_type(strategy.experience)
            self.curr_table = Table(title=f"Eval Metrics - {stream} stream")
            self.is_table_initialized = False
            self.live = Live(self.curr_table, refresh_per_second=4)
            self.live.start()

    def after_training(self, strategy: 'BaseStrategy',
                       metric_values: List['MetricValue'], **kwargs):
        print('[bold blue]End of training phase', file=self.file, flush=True)

    def after_training_iteration(self, strategy: 'BaseStrategy',
                                 metric_values: List['MetricValue'], **kwargs):
        if self.interactive:
            desc = ""
            for mval in metric_values:
                desc += f" {mval.name.split('/')[0]}={mval.value:.4f};"
            self._progress.set_description(desc)
            self._progress.update()
            self._progress.refresh()

    def before_training_epoch(self, strategy: 'BaseStrategy',
                              metric_values: List['MetricValue'], **kwargs):
        if hasattr(strategy.dataloader, '__len__'):
            self._progress.total = len(strategy.dataloader)

    def after_eval(self, strategy: 'BaseStrategy',
                   metric_values: List['MetricValue'], **kwargs):
        self.live.stop()
        self.live = None

        print('\t[italic]-- stream metrics', file=self.file, flush=True)
        sorted_vals = sorted(metric_values, key=lambda x: x.name)
        for mval in sorted_vals:
            name, x, val = mval.name, mval.x_plot, mval.value
            if isinstance(val, TensorImage):
                str_name = '_'.join(name.split('/'))
                save_image(val.image, f'{self.logdir}/{str_name}_x_{x}.png')
                val = f'<serialized to {str_name}>.'
            else:
                val = self._val_to_str(val)
            name = name.split('/')[0]
            print(f'\t{name} = {val}', file=self.file, flush=True)
        print('[bold blue]End of eval phase', file=self.file, flush=True)
        self.metric_vals = {}

    @property
    def _progress(self):
        if self._pbar is None:
            self._pbar = tqdm(leave=True, position=0, mininterval=1, file=sys.stdout)
        return self._pbar

    def _end_progress(self):
        if self._pbar is not None:
            self._pbar.close()
            self._pbar = None
