from collections import defaultdict
from typing import Union, Dict

import torch
from torch import Tensor

from avalanche.evaluation import Metric, GenericPluginMetric
from avalanche.evaluation.metric_results import MetricValue
from avalanche.evaluation.metric_utils import get_metric_name
from avalanche.evaluation.metrics import Mean


class Accuracy(Metric[float]):
    """Accuracy metric.

    The "prediction" and "target" tensors may contain plain labels or
    one-hot/logit vectors.

    Each time `result` is called, this metric emits the average accuracy
    across all predictions made since the last `reset`.

    The reset method will bring the metric to its initial state. By default
    this metric in its initial state will return an accuracy value of 0.
    """

    def __init__(self):
        """Creates an instance of the standalone Accuracy metric."""
        self._mean_accuracy = Mean()

    @torch.no_grad()
    def update(
        self,
        predicted_y: Tensor,
        true_y: Tensor,
    ) -> None:
        """Update the running accuracy given the true and predicted labels.

        :param predicted_y: The model prediction. Both labels and logit vectors
            are supported.
        :param true_y: The ground truth. Both labels and one-hot vectors
            are supported.

        :return: None.
        """
        if len(true_y) != len(predicted_y):
            raise ValueError("Size mismatch for true_y and predicted_y tensors")

        true_y = torch.as_tensor(true_y)
        predicted_y = torch.as_tensor(predicted_y)

        # Check if logits or labels
        if len(predicted_y.shape) > 1:
            # Logits -> transform to labels
            predicted_y = torch.max(predicted_y, 1)[1]

        if len(true_y.shape) > 1:
            # Logits -> transform to labels
            true_y = torch.max(true_y, 1)[1]

        true_positives = float(torch.sum(torch.eq(predicted_y, true_y)))
        total_patterns = len(true_y)
        self._mean_accuracy.update(true_positives / total_patterns, total_patterns)

    def result(self) -> float:
        """Retrieves the running accuracy.

        Calling this method will not change the internal state of the metric.

        :return: current running accuracy.
        """
        return self._mean_accuracy.result()

    def reset(self) -> None:
        """Resets the metric.

        :return: None.
        """
        self._mean_accuracy.reset()


class AvgStreamAccuracyPM(GenericPluginMetric[float]):
    """
    Base class for all accuracies plugin metrics
    """

    def __init__(self):
        self._accuracy = Accuracy()
        super().__init__(self._accuracy, reset_at="stream", emit_at="stream", mode="eval")

    def update(self, strategy):
        self._accuracy.update(strategy.mb_output, strategy.mb_y)

    def __str__(self):
        return "AvgStreamAcc"

    def _package_result(self, strategy: "SupervisedTemplate") -> "MetricResult":
        metric_name = get_metric_name(
            self, strategy,
            add_experience=False,
            add_task=False
        )
        metric_value = self.result(strategy)
        plot_x_position = strategy.clock.train_iterations

        return [
            MetricValue(self, metric_name, metric_value, plot_x_position)
        ]