import os
import random

from PIL import Image
from torch.utils.data import Dataset
import torch

from avalanche.benchmarks.utils.dataset_utils import ConstantSequence
import torchvision.transforms as tfs
from dac.benchmarks.single_image_data import MyRandomResizedCrop
import math


class AugmentedDataset(torch.utils.data.Dataset):
    def __init__(self, data, initcrop=0.5, degrees=30, scale=(0.01, 1.0), shear=30, vflip=False,
                 crop_size=256, mean=(0, 0, 0), std=(1, 1, 1), randinterp=False, cropfirst=True, debug=False,
                 length=int(1.281e6)):
        """Dataset class that combines the following transformations to generate samples from A SINGLE IMAGE

        1. RandomCrop (initialcrop) if cropfirst
        2. Random-Resized-Crop  if scale[0] > 0
            * size=(sqrt(2)*cropsize)            # sqrt(2) in the case of 45deg rotation.
            * scale=(1/scale[0], 1/scale[1])
        3. RandomAffine (degrees and shear) if shear!=0 and degrees!=0
        4. VerticalFlip (p=0.5) if vflip
        5. RandomHorizontalflip (p=0.5)
        6. CenterCrop (cropsize)
        7. ColorJitter
        #  8. RandomGaussianBlur # from MoCo-v2 repo
        9. ...
        """
        self.data = data

        normalize = tfs.Normalize(mean=mean, std=std)
        scale = (1. / scale[0], 1. / scale[1])
        if randinterp and not debug:
            resizedcrop = tfs.RandomApply(
                [MyRandomResizedCrop(int(math.sqrt(2) * crop_size), scale=scale, interpolation=1),
                 MyRandomResizedCrop(int(math.sqrt(2) * crop_size), scale=scale, interpolation=2),
                 MyRandomResizedCrop(int(math.sqrt(2) * crop_size), scale=scale, interpolation=3)])
        else:
            resizedcrop = MyRandomResizedCrop(int(math.sqrt(2) * crop_size), scale=scale, interpolation=3)
        self.length = length  # run out after one IN-1k epoch
        self.crop_size = crop_size
        self.totensor = tfs.ToTensor()

        tfslist = []
        if cropfirst:
            tfslist.append(tfs.RandomCrop(int(initcrop * min(self.img.size))))

        if scale[0] > 0:
            tfslist.append(resizedcrop)
        if shear != 0:
            tfslist.append(tfs.RandomAffine(degrees, translate=None, scale=None, shear=shear, resample=Image.BILINEAR,
                                            fillcolor=0))
        elif degrees != 0:
            tfslist.append(tfs.RandomRotation(degrees=degrees, resample=Image.BILINEAR))

        if vflip:
            tfslist.append(tfs.RandomVerticalFlip(p=0.5))

        tfslist.append(tfs.CenterCrop(crop_size))
        if debug:
            tfslist.append(tfs.Resize((32, 32)))
        else:
            tfslist.extend([
                tfs.RandomApply([tfs.ColorJitter(brightness=0.4, contrast=0.4, saturation=0.4, hue=0.1)], 0.5),
                # GaussianBlur()
            ])
        tfslist.append(tfs.ToTensor())
        if mean != [0, 0, 0]:
            tfslist.append(normalize)
        self.transforms = tfs.Compose(tfslist)
        print("transforms:", tfslist)

    def __len__(self):
        return self.length

    def __getitem__(self, idx):
        rand_idx = random.randint(0, len(self.data) - 1)
        img = self.data[rand_idx]
        return self.transforms(img)


class DistillationDatasetWrapped(Dataset):
    def __init__(self, data, transform):
        """DistillationDataset.

        Images are stored in a single folder. No subfolders.

        :param root:
        :param transform:
        """
        self.orig_data = data
        self.transform = transform
        self.targets = ConstantSequence(0, len(self.orig_data))

    def __len__(self):
        return len(self.orig_data)

    def __getitem__(self, idx):
        img = self.orig_data[idx]
        with torch.no_grad():
            tensor_image = self.transform(img)
        return tensor_image, 0


class DistillationDatasetFromDir(Dataset):
    def __init__(self, root, transform=None):
        """DistillationDataset.

        Images are stored in a single folder. No subfolders.

        :param root:
        :param transform:
        """
        self.root = root
        self.transform = transform
        self.total_imgs = os.listdir(root)
        # self.targets = ConstantSequence(0, len(self.total_imgs))

    def __len__(self):
        return len(self.total_imgs)

    def __getitem__(self, idx):
        img_loc = os.path.join(self.root, self.total_imgs[idx])
        image = Image.open(img_loc).convert("RGB")
        if self.transform is not None:
            with torch.no_grad():
                image = self.transform(image)
        return image


class SyntheticDatasetFromDir(Dataset):
    def __init__(self, root, transform=None):
        """Synthetic datasets have the follo

        Images are stored in a single folder. No subfolders.

        :param root:
        :param transform:
        """
        self.root = root
        self.transform = transform

        self.total_imgs = []
        for classdir in os.listdir(root):
            classimgs = os.listdir(os.path.join(root, classdir))
            self.total_imgs.extend([os.path.join(classdir, img) for img in classimgs])
        # self.targets = ConstantSequence(0, len(self.total_imgs))

    def __len__(self):
        return len(self.total_imgs)

    def __getitem__(self, idx):
        img_loc = os.path.join(self.root, self.total_imgs[idx])
        image = Image.open(img_loc).convert("RGB")
        if self.transform is not None:
            with torch.no_grad():
                image = self.transform(image)
        return image