import random

import torch
import torch.utils
from hydra.utils import instantiate
from omegaconf import DictConfig, OmegaConf
import ray
import os

from torch.utils.data import Subset
from torchvision import transforms

from avalanche.benchmarks.utils import AvalancheDataset
from dac.utils import relative_gpu_id, get_best_gpus
from dac.ood_data import DistillationDatasetWrapped


def train_preamble(cfg: DictConfig):
    if torch.cuda.is_available():
        # if ray.is_initialized():
        #     gpuid = relative_gpu_id(ray.get_gpu_ids()[0])
        #     device = f'cuda:{gpuid}'
        #     torch.cuda.set_device(device)
        #     print("train_config using GPU device: ", device)
        # else:
        # if not using ray, assume CUDA_VISIBLE_DEVICES is configured correctly.
        # else:
        #     gpuid = relative_gpu_id(get_best_gpus(1)[0])
        device = 'cuda'
    else:
        device = "cpu"
        print("WARNING: NO CUDA. Using CPU!!!!")
        print("WARNING: NO CUDA. Using CPU!!!!")
        print("WARNING: NO CUDA. Using CPU!!!!")
        print("WARNING: NO CUDA. Using CPU!!!!")
        print("WARNING: NO CUDA. Using CPU!!!!")

    if cfg.experiment_type == "model_assessment":
        basedir = os.getcwd()
        rundir = f"run{cfg.run_id}"
        if rundir not in basedir:
            basedir = os.path.join(basedir, rundir)
            os.makedirs(basedir, exist_ok=True)
            os.chdir(basedir)
    else:
        basedir = os.getcwd()

    print(cfg)
    print("Working directory : {}".format(basedir))
    OmegaConf.save(cfg, f"{basedir}/hydra_config.yaml")
    return device, basedir


def get_online_mean_std_normalizer(data, max_count=100):
    mean, std = 0, 0
    count = 0
    for batch in torch.utils.data.DataLoader(
            data, batch_size=32, shuffle=False, num_workers=8):
        mean += batch.mean(dim=(0, 2, 3))
        std += batch.std(dim=(0, 2, 3))
        count += 1
        if count >= max_count:
            break
    return mean / count, std / count


class RandomSubset:
    def __init__(self, data, length):
        self.data = data
        self.length = length

    def __getitem__(self, item):
        rand_idx = random.randint(0, len(self.data) - 1)
        return self.data[rand_idx]

    def __len__(self):
        return self.length


def get_ood_data(cfg: DictConfig):
    crop_size = cfg.benchmark.img_resolution

    if cfg.debug_mode:
        if 'crop_size' in cfg.ood_data.ood_data:
            ood_data = instantiate(cfg.ood_data.ood_data, crop_size=crop_size)
        else:
            ood_data = instantiate(cfg.ood_data.ood_data)
        ood_data = Subset(ood_data, range(1000))
    else:
        if 'crop_size' in cfg.ood_data.ood_data:
            ood_data = instantiate(cfg.ood_data.ood_data, crop_size=crop_size)
        else:
            ood_data = instantiate(cfg.ood_data.ood_data)

    if len(ood_data) > 100000:
        print("******* subsampling data: *********\n", cfg.ood_data)
        ood_data = RandomSubset(ood_data, 50000)

    print("LENGTH: ", len(ood_data))

    # sanity check: image resolution correspond to original benchmark
    for batch in torch.utils.data.DataLoader(ood_data):
        assert batch[0].shape[-1] >= crop_size

    MEAN, STD = get_online_mean_std_normalizer(ood_data)
    print(f"MEAN={MEAN}, STD={STD}")
    # STD = (0.2023, 0.1994, 0.2010)  # imagenet statistics
    # MEAN = (0.4914, 0.4822, 0.4465)
    transform = transforms.Compose([
        transforms.RandomCrop(crop_size, padding=4),
        transforms.RandomHorizontalFlip(),
        transforms.Normalize(MEAN, STD),
    ])
    train_data = DistillationDatasetWrapped(ood_data, transform)
    return AvalancheDataset(train_data)
