"""Train in macl scenarios."""
import os

import torch
from omegaconf import DictConfig
from hydra.utils import instantiate
from ray import tune

from avalanche.models import as_multitask
from avalanche.training import Naive
from dac.utils import default_evaluator, to_json
from dac.training.configs.common import train_preamble, get_ood_data
from torch.utils.data import DataLoader, Subset
from torchvision.utils import save_image, make_grid

from dac.training.latent_kd_projection import LatentKDProjection
from dac.training.indipendent_oodkd import INDIP_OOD_KDStrategy


def train_config(cfg: DictConfig, base_dir=None):
    device, basedir = train_preamble(cfg)
    if base_dir is not None:
        basedir = base_dir
    print("BASEDIR", basedir)

    ########################################
    # SCENARIO PREPARATION
    ########################################
    macl_scenario = instantiate(cfg.benchmark.macl_benchmark_instance, run_id=cfg.run_id)
    original_scenario = instantiate(cfg.benchmark.seq_benchmark_instance, run_id=cfg.run_id)
    # for seq_exp, macl_exp in zip(original_scenario.train_stream, macl_scenario.expert_models_stream):
    #     strat = Naive(macl_exp.expert_model, None, eval_mb_size=128, device=device)
    #     strat.eval(seq_exp, num_workers=8, pin_memory=False)

    ood_data = get_ood_data(cfg)
    imgs_ood_sample = next(iter(DataLoader(ood_data, batch_size=32)))[0]
    save_image(make_grid(imgs_ood_sample, nrow=8), os.path.join(basedir, f"{basedir}/ood_sample.png"))

    ########################################
    # STRATEGY PREPARATION
    ########################################
    model = instantiate(cfg.benchmark.cl_model).to(device)
    model = as_multitask(model, cfg.benchmark.classifier_name)
    optimizer = instantiate(cfg.optimizer, model.parameters())

    plugs = [LatentKDProjection(
        latent_layers_infos=cfg.strategy.latent_layers_infos,
        lamb=cfg.strategy.latent_proj_lambda,
        proj_init=cfg.strategy.latent_proj_init)]
    evalp = default_evaluator(basedir, is_interactive=False)

    if cfg.debug_mode:
        print("DEBUG mode.")
        cfg.strategy.train_epochs = 3

    strategy = INDIP_OOD_KDStrategy(
        # KD arguments
        temperature=cfg.strategy.temperature,
        ood_dataset=ood_data,
        # BaseSGD arguments
        train_epochs=cfg.strategy.train_epochs,
        model=model,
        optimizer=optimizer,
        train_mb_size=cfg.strategy.train_mb_size,
        eval_mb_size=512,
        device=device,
        eval_every=50,
        plugins=list(cfg.strategy.plugins) + plugs,
        evaluator=evalp,
    )

    ########################################
    # TRAINING
    ########################################
    for i, train_exp in enumerate(macl_scenario.expert_models_stream):
        train_exp.task_labels = [i]
        train_exp.task_label = i

        evals = [original_scenario.train_stream]
        strategy.train(experiences=train_exp, eval_streams=evals, num_workers=8)

        eid = train_exp.current_experience
        model_fname = os.path.join(basedir, f'model_e{eid}.pth')
        torch.save(strategy.model.state_dict(), model_fname)

        if cfg.experiment_type == "model_selection":  # Report to Ray Tune during model selection
            mname = "AvgStreamAcc/eval_phase/train_stream"
            streamacc = strategy.evaluator.get_all_metrics()[mname][1][-1]
            tune.report(train_stream_acc=streamacc)

    strategy.eval(original_scenario.test_stream, num_workers=8)

    with open(os.path.join(basedir, 'metrics.json'), 'w') as f:
        f.write(to_json(strategy.evaluator.get_all_metrics(), ignore_errors=True))
