from torch.utils.data import DataLoader

from runs.train_ood_kd_ci import CI_OOD_KDStrategy
from dac.training.latent_kd_projection import LatentKDProjection

import os

import torch.optim
from hydra.utils import instantiate
from omegaconf import DictConfig
from ray import tune
from avalanche.models import as_multitask
from dac.utils import to_json, default_evaluator
from dac.training.configs.common import train_preamble
from torchvision.utils import make_grid, save_image

from dac.training.configs.common import get_ood_data


def train_config(cfg: DictConfig, base_dir=None):
    device, basedir = train_preamble(cfg)
    if base_dir is not None:
        basedir = base_dir
    print("BASEDIR", basedir)

    ########################################
    # SCENARIO PREPARATION
    ########################################
    scenario = instantiate(cfg.benchmark.seq_benchmark_instance, run_id=cfg.run_id)
    original_test_stream = scenario.test_stream
    original_train_stream = scenario.train_stream
    ood_data = get_ood_data(cfg)
    imgs_ood_sample = next(iter(DataLoader(ood_data, batch_size=32)))[0]
    save_image(make_grid(imgs_ood_sample, nrow=8), os.path.join(basedir, f"{basedir}/ood_sample.png"))

    ########################################
    # STRATEGY PREPARATION
    ########################################
    model = instantiate(cfg.benchmark.cl_model).to(device)
    model = as_multitask(model, cfg.benchmark.classifier_name)
    optimizer = instantiate(cfg.optimizer, model.parameters())
    plugs = [LatentKDProjection(
        latent_layers_infos=cfg.strategy.latent_layers_infos,
        lamb=cfg.strategy.latent_proj_lambda,
        proj_init=cfg.strategy.latent_proj_init)]
    evalp = default_evaluator(basedir, is_interactive=False)

    if cfg.debug_mode:
        print("DEBUG mode.")
        cfg.strategy.train_epochs = 3
        expert_epochs = 2
        original_train_stream = list(original_train_stream)[:3]
        original_test_stream = list(original_test_stream)[:3]
    else:
        expert_epochs = 200

    strategy = CI_OOD_KDStrategy(
        # KD arguments
        temperature=cfg.strategy.temperature,
        ood_dataset=ood_data,
        expert_epochs=expert_epochs,
        train_epochs=cfg.strategy.train_epochs,
        init_from_first=cfg.strategy.init_from_first,
        basedir=basedir,
        # BaseSGD arguments
        model=model,
        optimizer=optimizer,
        train_mb_size=512,
        eval_mb_size=512,
        device=device,
        eval_every=50,
        plugins=list(cfg.strategy.plugins) + plugs,
        evaluator=evalp,
        classifier_name=cfg.benchmark.classifier_name,
    )

    ########################################
    # TRAINING
    ########################################
    for i, train_exp in enumerate(original_train_stream):
        evals = [original_train_stream]
        strategy.train(experiences=train_exp, eval_streams=evals, num_workers=8, pin_memory=False)
        strategy.eval(original_test_stream, num_workers=8, pin_memory=False)

        eid = train_exp.current_experience
        torch.save(strategy.model.state_dict(), f'{basedir}/model_e{eid}.pth')
        torch.save(strategy.expert_model.state_dict(), f'{basedir}/expert_e{eid}.pth')

        if cfg.experiment_type == "model_selection":  # Report to Ray Tune during model selection
            mname = "AvgStreamAcc/eval_phase/train_stream"
            streamacc = strategy.evaluator.get_all_metrics()[mname][1][-1]
            tune.report(train_stream_acc=streamacc)

    strategy.eval(original_test_stream, num_workers=8)
    with open(os.path.join(f'{basedir}/metrics.json'), 'w') as f:
        f.write(to_json(strategy.evaluator.get_all_metrics(), ignore_errors=True))
