import copy

import torch

from avalanche.evaluation.metric_results import MetricValue
from avalanche.models import avalanche_forward
from avalanche.training import Naive
from dac.cutmix import CutMix
from dac.training.losses import masked_kd_loss
import torch.nn.functional as F


class INDIP_OOD_KDStrategy(Naive):
    def __init__(self, temperature, ood_dataset, **kwargs):
        super().__init__(**kwargs)
        self.temperature = temperature
        self.ood_dataset = ood_dataset

        self._cm = CutMix()

        # Strategy state
        self.expert_model = None
        self.prev_model = None

    def _before_training_exp(self, **kwargs):
        self.expert_model = self.experience.expert_model
        self.expert_model = self.expert_model.to(self.device).eval()
        super()._before_training_exp(**kwargs)

    def train_dataset_adaptation(self, **kwargs):
        self.adapted_dataset = self.ood_dataset

    def _before_forward(self, **kwargs):
        # apply CutMix during training.
        if self.is_training:
            mb_x, _, _, _ = self._cm(self.mb_x, self.mb_y)
            self.mbatch[0] = mb_x
        super()._before_forward(**kwargs)

    def _train_criterion(self):
        curr_tid = self.experience.task_label
        y_student = self.model(self.mb_x, task_labels=None)  # all heads

        if self.prev_model is not None:
            with torch.no_grad():  # all heads for prev. CL model
                y_teacher_cl = avalanche_forward(self.prev_model, self.mb_x, None)
            assert len(y_teacher_cl) + 1 == len(y_student)  # check number of heads
            # compute loss over all the previous heads
            kd_cl = 0.0
            for hid in y_teacher_cl:
                kd_cl += masked_kd_loss(y_student[hid], y_teacher_cl[hid], temperature=self.temperature)
        else:
            kd_cl = torch.tensor(0.0)

        with torch.no_grad():  # expert model is single task, prev model is multi-task
            y_teacher_expert = self.expert_model(self.mb_x)
        kd_expert = masked_kd_loss(y_student[curr_tid], y_teacher_expert, temperature=self.temperature)

        mval = MetricValue(self, 'kd_cl', kd_cl.detach().cpu().item(), self.clock.train_iterations)
        self.evaluator.publish_metric_value(mval)
        mval = MetricValue(self, 'kd_expert', kd_expert.detach().cpu().item(), self.clock.train_iterations)
        self.evaluator.publish_metric_value(mval)

        return kd_cl + kd_expert  # no need to balance since heads are summed

    def criterion(self):
        if self.is_training:
            return self._train_criterion()
        else:
            return F.cross_entropy(self.mb_output, self.mb_y)

    def _after_training_exp(self, **kwargs):
        super()._after_training_exp(**kwargs)
        self.prev_model = copy.deepcopy(self.model)
        self.prev_model.eval()

