from avalanche.evaluation.metric_results import MetricValue

from avalanche.models.helper_method import MultiTaskDecorator
from functools import reduce
import torch.nn.functional as F

import torch

from avalanche.core import BaseSGDPlugin


class LatentKDProjection(BaseSGDPlugin):
    def __init__(self, latent_layers_infos, lamb, proj_init='eye', basedir='.'):
        """Latent KD Projection.

        :param latent_layers_infos: list of triplets <module name, use-output, hidden-size>
        :param lamb:
        """
        super().__init__()
        self.latent_layers_infos = latent_layers_infos
        self.lamb = lamb
        self.basedir = basedir

        assert proj_init in {'eye', 'xavier_uniform'}
        self.proj_init = proj_init

        self._handles = []
        self.curr_exp_acts = []
        self.prev_exp_acts = []
        self.cl_acts = []
        self._projection_matrices = torch.nn.ParameterDict()  # an empty container to hold projection matrices
        self._proj_optimizer = None

    def _get_module(self, model, lname):
        if isinstance(model, MultiTaskDecorator):
            lname = 'model.' + lname
        return reduce(getattr, lname.split('.'), model)

    def _make_model_hooks(self, model_act_list, model):
        """Add forward hooks to collect activations."""
        def hook_with_list(act_list_name, use_output, module, input, output):
            nonlocal self
            if use_output:
                getattr(self, act_list_name).append(output)
            else:
                getattr(self, act_list_name).append(input[0])

        for lname, use_output, _ in self.latent_layers_infos:
            fh = lambda m, i, o, uo=use_output: hook_with_list(model_act_list, uo, m, i, o)
            mod = self._get_module(model, lname)
            handle = mod.register_forward_hook(fh)
            self._handles.append(handle)

    def _make_projection_matrix(self, model_name):
        for ii, (lname, uo, lsize) in enumerate(self.latent_layers_infos):
            if self.proj_init == 'eye':
                P = torch.nn.Parameter(torch.eye(lsize))
            elif self.proj_init == 'xavier_uniform':
                w = torch.empty(lsize, lsize)
                torch.nn.init.xavier_uniform_(w)
                P = torch.nn.Parameter(w)
            else:
                assert False

            k = f"{model_name}_{lname.replace('.', '-')}"
            self._projection_matrices[k] = P

    def before_training_exp(self, strategy, **kwargs):
        """Add forward hooks to student and experts."""
        assert len(self._handles) == 0
        # forward hooks
        self._make_model_hooks('cl_acts', strategy.model)
        self._make_model_hooks('curr_exp_acts', strategy.expert_model)
        if strategy.prev_model is not None:
            self._make_model_hooks('prev_exp_acts', strategy.prev_model)
        # projection matrices
        self._make_projection_matrix('curr_expert')
        if strategy.prev_model is not None:
            self._make_projection_matrix('prev_expert')
        self._projection_matrices.to(strategy.device)
        self._proj_optimizer = torch.optim.SGD(self._projection_matrices.parameters(), lr=0.0001)

    def latent_kd_proj_loss(self, strategy, expert_acts_list, model_name):
        proj_loss = 0
        for ii, (lname, _, _) in enumerate(self.latent_layers_infos):
            P = getattr(self._projection_matrices, f"{model_name}_{lname.replace('.', '-')}")
            h_new = self.cl_acts[ii].reshape(self.cl_acts[ii].shape[0], -1)
            h_proj = torch.matmul(h_new, P)
            h_expert = expert_acts_list[ii].reshape(expert_acts_list[ii].shape[0], -1)
            ll = F.mse_loss(h_proj, h_expert)

            mname = f"proj_loss/{model_name}_{lname.replace('.', '-')}"
            mval = MetricValue(self, mname, ll.detach().cpu().item(),
                               strategy.clock.train_iterations)
            strategy.evaluator.publish_metric_value(mval)

            proj_loss += ll
        return proj_loss

    def before_forward(self, strategy, **kwargs):
        self.curr_exp_acts = []
        self.prev_exp_acts = []
        self.cl_acts = []

    def before_backward(self, strategy, **kwargs):
        """Latent KD projection loss. Reset activation lists at the end."""
        assert len(self.curr_exp_acts) == len(self.latent_layers_infos)
        numexp = strategy.clock.train_exp_counter

        l_curr = self.latent_kd_proj_loss(strategy, self.curr_exp_acts, 'curr_expert')
        if strategy.prev_model is not None:
            l_prev = self.latent_kd_proj_loss(strategy, self.prev_exp_acts, 'prev_expert')
        else:
            l_prev = 0
        strategy.loss += self.lamb * (l_curr + numexp * l_prev)

    def after_backward(self, strategy, *args, **kwargs):
        """Descent step for projection matrices"""
        self._proj_optimizer.step()
        self._proj_optimizer.zero_grad()

    def after_training_exp(self, strategy, **kwargs):
        """remove forward hooks and projection matrices."""
        for h in self._handles:
            h.remove()
        self._handles = []
        eid = strategy.clock.train_exp_counter
        for name, P in self._projection_matrices.items():
            torch.save(P.data, f"{self.basedir}/P{name}_e{eid}.pt")
        self._projection_matrices = torch.nn.ParameterDict()

