from avalanche.core import SupervisedPlugin
import torch
import os
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from avalanche.training import BaseStrategy


class ModelCheckpoint(SupervisedPlugin):
    """Model checkpoint plugin that saves the best model by tracking a metric
    and load it at the end of training."""

    def __init__(self, metric_name='Top1_Acc_Stream/eval_phase/valid_stream/Task000',
                 log_dir='.'):
        """Init.

        :param metric_name: name of the metric to keep track of.
        :param log_dir: name of the directory.
        """
        super().__init__()
        self.metric_name = metric_name
        self.best_val = -10000
        self.log_dir = log_dir
        self.model_fname = os.path.join(self.log_dir, 'best_model.pt')

    def after_eval_exp(self, strategy: 'BaseStrategy', **kwargs):
        self.maybe_save_model(strategy)

    def after_training_exp(self, strategy: 'BaseStrategy', **kwargs):
        vals = strategy.evaluator.last_metric_results
        if self.metric_name not in vals:
            print("ERROR: ModelCheckpoint is tracking a metric that is not computed.")
            assert False

        strategy.model = torch.load(self.model_fname)

    def maybe_save_model(self, strategy):
        print("model checkpoint.")
        vals = strategy.evaluator.last_metric_results
        if self.metric_name in vals:
            curr_val = vals[self.metric_name]
            if curr_val > self.best_val:
                print("New best value. Saving model checkpoint...")
                self.best_val = curr_val
                torch.save(strategy.model, self.model_fname)
        else:
            return
