import os

import random
from types import SimpleNamespace

import numpy
import numpy as np
import torch
from avalanche.evaluation.metrics import accuracy_metrics, loss_metrics, \
    timing_metrics, forgetting_metrics
from avalanche.logging import TextLogger, TensorboardLogger
from avalanche.training import Naive
from avalanche.training.plugins import EvaluationPlugin
from avalanche.benchmarks.classic.ccifar100 import _get_cifar100_dataset, _default_cifar100_train_transform, \
    _default_cifar100_eval_transform

from avalanche.benchmarks import nc_benchmark
from avalanche.benchmarks.utils import AvalancheDataset
from avalanche.benchmarks import CLStream
from dac.evaluation.task_agnostic_accuracy import AvgStreamAccuracyPM
from dac.evaluation.rich_text_logging import RichTextLogger
from torch import nn

from dac.training.macl_data_buffers import _default_get_classifier


SEED_BENCHMARK_RUNS = [1234, 2345, 3456, 5678, 6789]  # 5 different seeds to randomize class orders


def plotstyle_iclr2022(*, usetex=True, rel_width=1.0, nrows=1, ncols=1, family="serif"):
    """Neurips 2022 bundle."""
    from tueplots import fonts, figsizes, fontsizes
    if usetex is True:
        font_config = fonts.neurips2022_tex(family=family)
    elif usetex is False:
        font_config = fonts.neurips2022(family=family)
    size = figsizes.neurips2022(rel_width=rel_width, nrows=nrows, ncols=ncols)
    fontsize_config = fontsizes._from_base(base=12)
    return {**font_config, **size, **fontsize_config}


def relative_gpu_id(global_gpu_id: int):
    global_gpu_id = int(global_gpu_id)
    visible_ids = os.environ["CUDA_VISIBLE_DEVICES"]
    gpu_ids = [int(el) for el in visible_ids.split(',')]
    for ridx, gidx in enumerate(gpu_ids):
        if global_gpu_id == gidx:
            return ridx
    raise ValueError(f"Unkown global_gpu_id. global_gpu_id={global_gpu_id}"
                     f" CUDA_VISIBLE_DEVICES={visible_ids}")


class GaussianNoise(object):
    def __init__(self, mean=0., std=1.):
        """ pytorch transform. for additive gaussian noise. """
        self.std = std
        self.mean = mean

    def __call__(self, tensor):
        noise = torch.randn(tensor.size(), device=tensor.device)
        return tensor + noise * self.std + self.mean

    def __repr__(self):
        return self.__class__.__name__ + \
               '(mean={0}, std={1})'.format(self.mean, self.std)


def eval_model_accuracy(model, stream):
    """Eval model on data

    :args model: model to evaluate
    :args stream: data to evaluate on. Can be AvalancheDataset or list of AvalancheDatasets.
    """
    if isinstance(stream, AvalancheDataset):
        stream = SimpleNamespace(
            dataset=stream,
            task_label=0,
            task_labels=[0],
            classes_in_this_experience=list(set(stream.targets))
        )
        stream = CLStream(name="train", exps_iter=[stream])

    avgacc = AvgStreamAccuracyPM()

    evalp = EvaluationPlugin(avgacc, loggers=[])
    strat = Naive(model, None, eval_mb_size=512,
                  device='cuda', evaluator=evalp)
    strat.eval(list(stream), num_workers=8)
    return avgacc._metric.result()


def default_evaluator(log_dir, is_interactive=True):
    text_logger = TextLogger(open(os.path.join(log_dir, 'log.txt'), 'w'))
    tb_logger = TensorboardLogger(os.path.join(log_dir, 'tb'))
    rtl = RichTextLogger(log_dir, interactive=is_interactive)
    return EvaluationPlugin(
        accuracy_metrics(epoch=True, experience=True, stream=True, minibatch=True),
        loss_metrics(epoch=True, experience=True, stream=True),
        timing_metrics(epoch=True),
        forgetting_metrics(experience=True, stream=True),
        AvgStreamAccuracyPM(),
        loggers=[text_logger, tb_logger, rtl])


def in_notebook():
    try:
        from IPython import get_ipython
        if 'IPKernelApp' not in get_ipython().config:  # pragma: no cover
            return False
    except ImportError:
        return False
    return True


class UnNormalize(object):
    def __init__(self, mean, std):
        self.mean = mean
        self.std = std

    def __call__(self, tensor):
        """
        Args:
            tensor (Tensor):
                Tensor image of size (C, H, W) or batch of images (B, C, H, W)
                to be normalized.
        Returns:
            Tensor: Normalized image.
        """
        m = torch.tensor(self.mean, device=tensor.device).reshape(-1, 1, 1)
        s = torch.tensor(self.std, device=tensor.device).reshape(-1, 1, 1)
        tensor.mul_(s).add_(m)
        return tensor


def make_list(elements):
    if elements[0] == 'None':
        return []
    return list(elements)


def get_classifier_weights_from_args(args):
    gc = _default_get_classifier
    if 'resnet' in args.model:
        gc = lambda m: m.linear.weight
    elif 'mobile' in args.model:
        gc = lambda m: m.classifier[1].weight
    elif 'lenet' in args.model:
        gc = lambda m: m.classifier[0].weight
    return gc


def get_classifier_from_args(args):
    gc = _default_get_classifier
    if 'resnet' in args.model:
        gc = lambda m: m.linear
    elif 'mobile' in args.model:
        gc = lambda m: m.classifier[1]
    elif 'lenet' in args.model:
        gc = lambda m: m.classifier[0]
    return gc


class ModelClassifierWrapper(nn.Module):
    def __init__(self, model, get_classifier):
        super().__init__()
        self.model = model
        self.get_classifier = get_classifier

    @property
    def classifier(self):
        return self.get_classifier(self.model)

    def forward(self, x):
        return self.model(x)


def with_classifier_attribute(model, args):
    return ModelClassifierWrapper(model, get_classifier_weights_from_args(args))


INDENT = 3
SPACE = " "
NEWLINE = "\n"


def to_json(o, level=0, ignore_errors=False):
    """ pretty-print json.
    source: https://stackoverflow.com/questions/10097477/python-json-array-newlines
    :param o:
    :param level:
    :return:
    """
    # TODO: fix NaN encoding from 'nan' to 'NaN'. 'nan' is non-standard and crashes during loading.
    ret = ""
    if isinstance(o, dict):
        ret += "{" + NEWLINE
        comma = ""
        for k, v in o.items():
            ret += comma
            comma = ",\n"
            ret += SPACE * INDENT * (level + 1)
            ret += '"' + str(k) + '":' + SPACE
            ret += to_json(v, level + 1, ignore_errors)

        ret += NEWLINE + SPACE * INDENT * level + "}"
    elif isinstance(o, str):
        ret += '"' + o + '"'
    elif isinstance(o, list):
        ret += "[" + ",".join([to_json(e, level + 1, ignore_errors) for e in o]) + "]"
    # Tuples are interpreted as lists
    elif isinstance(o, tuple):
        ret += "[" + ",".join(to_json(e, level + 1, ignore_errors) for e in o) + "]"
    elif isinstance(o, bool):
        ret += "true" if o else "false"
    elif isinstance(o, int):
        ret += str(o)
    elif isinstance(o, float):
        ret += '%.7g' % o
    elif isinstance(o, numpy.ndarray) and numpy.issubdtype(o.dtype, numpy.integer):
        ret += "[" + ','.join(map(str, o.flatten().tolist())) + "]"
    elif isinstance(o, numpy.ndarray) and numpy.issubdtype(o.dtype, numpy.inexact):
        ret += "[" + ','.join(map(lambda x: '%.7g' % x, o.flatten().tolist())) + "]"
    elif o is None:
        ret += 'null'
    elif ignore_errors:
        # we do not recognize the type but we don't want to raise an error.
        ret = '"<not serializable>"'
    else:
        # Unknown type. Raise error.
        raise TypeError("Unknown type '%s' for json serialization" % str(type(o)))
    return ret


def get_best_gpus(num_gpus):
    try:
        import gpustat
    except ImportError:
        print("gpustat module is not installed. No GPU allocated.")
        return []

    try:
        selected = []
        stats = gpustat.GPUStatCollection.new_query()

        for i in range(num_gpus):
            ids_mem = [res for res in map(lambda gpu: (int(gpu.entry['index']),
                                          float(gpu.entry['memory.used']) /\
                                          float(gpu.entry['memory.total'])),
                                      stats) if str(res[0]) not in selected]

            if len(ids_mem) == 0:
                # No more gpus available
                break

            best = min(ids_mem, key=lambda x: x[1])
            bestGPU, bestMem = best[0], best[1]
            # print(f"{i}-th best GPU is {bestGPU} with mem {bestMem}")
            selected.append(str(bestGPU))

        print("Using GPUs: {}".format(",".join(selected)))
        return selected
    except BaseException as e:
        print("GPU not available: " + str(e))


def maybe_set_gpus(num_gpus):
    if 'CUDA_VISIBLE_DEVICES' in os.environ:
        print("CUDA_VISIBLE_DEVICES is already set. Not changing it.")
        return
    try:
        selected = get_best_gpus(num_gpus)
        print("Setting CUDA_VISIBLE_DEVICES GPUs to: {}".format(",".join(selected)))
        os.environ['CUDA_DEVICE_ORDER'] = 'PCI_BUS_ID'
        os.environ['CUDA_VISIBLE_DEVICES'] = ",".join(selected)
        return selected
    except BaseException as e:
        print("GPU not available: " + str(e))


def init_deterministic_run(seed):
    """Sets the libraries to use deterministic computations.

    Use this at the top of your main script.
    """
    torch.manual_seed(seed)
    torch.use_deterministic_algorithms(True)
    random.seed(seed)
    np.random.seed(seed)


def get_mt_cifar100(seed):
    cifar_train, cifar_test = _get_cifar100_dataset(None)

    return nc_benchmark(
        train_dataset=cifar_train,
        test_dataset=cifar_test,
        n_experiences=10,
        task_labels=True,
        seed=seed,
        shuffle=True,
        class_ids_from_zero_in_each_exp=False,
        train_transform=_default_cifar100_train_transform,
        eval_transform=_default_cifar100_eval_transform,
    )
