import copy
import os

import argparse
import torch.optim
from torch.nn import CrossEntropyLoss
from torch.utils.data import DataLoader
from torchvision import transforms
from torch import nn
from avalanche.benchmarks.utils import AvalancheClassificationDataset, AvalancheConcatClassificationDataset
from avalanche.benchmarks import SplitCIFAR10
from avalanche.evaluation.metric_results import MetricValue
from dac.cutmix import CutMix
from dac.gem_resnet import ResNet18

from avalanche.training import Naive
from avalanche.models import MultiTaskModule
from dac.ood_data import DistillationDatasetFromDir
from dac.utils import to_json
import torch.nn.functional as F


parser = argparse.ArgumentParser(description="Knowledge Distillation From a Single Image.")
parser.add_argument("--batch_size",
                    default=64, type=int,
                    help="batch size")
parser.add_argument("--learning_rate",
                    default=0.001, type=float,
                    help="initial learning rate")
parser.add_argument("--temperature",
                    default=2, type=float,
                    help="temperature for distillation")


class OODKDStrategy(Naive):
    def __init__(self, alpha, temperature, ood_dataset, **kwargs):
        super().__init__(**kwargs)
        self.alpha = alpha
        self.temperature = temperature
        self.ood_dataset = ood_dataset

        self.ood_it = iter(DataLoader(
            self.ood_dataset,
            batch_size=self.train_mb_size,
            num_workers=8,
            shuffle=True))

        self.eval_criterion = nn.CrossEntropyLoss()
        self.train_criterion = nn.MSELoss()
        self.cm = CutMix()

        # Strategy state
        self.logits_target = None
        self.prev_model = None
        self.prev_classes = []

    def _before_backward(self, **kwargs):
        """
        Add distillation loss
        """
        alpha = (
            self.alpha[self.clock.train_exp_counter]
            if isinstance(self.alpha, (list, tuple))
            else self.alpha
        )
        ood_x = self.next_ood_batch()[0]
        ood_pred = self.model(ood_x)
        penalty = self._penalty(ood_pred, ood_x, alpha)

        mval = MetricValue(self, 'ce', self.loss.detach().cpu().item(), self.clock.train_iterations)
        self.evaluator.publish_metric_value(mval)
        mval = MetricValue(self, 'kd', penalty.detach().cpu().item(), self.clock.train_iterations)
        self.evaluator.publish_metric_value(mval)
        self.loss += penalty

    def _distillation_loss(self, curr_out, prev_out, active_units):
        """
        Compute distillation loss between output of the current model and
        and output of the previous (saved) model.
        """
        # we compute the loss only on the previously active units.
        # au = list(active_units)
        # curr_out = curr_out[:, au] / self.temperature
        # prev_out = prev_out[:, au] / self.temperature

        curr_out = curr_out / self.temperature
        prev_out = prev_out / self.temperature

        res = F.kl_div(
            F.log_softmax(curr_out, dim=1),
            F.log_softmax(prev_out, dim=1),
            log_target=True,
            reduction="batchmean")

        return res

    def _penalty(self, student_output, x, alpha):
        """
        Compute weighted distillation loss.
        """

        if self.prev_model is None:
            return torch.tensor(0)
        else:
            with torch.no_grad():
                assert not isinstance(self.prev_model, MultiTaskModule)
                y_prev = self.prev_model(x)
            y_curr = student_output
            dist_loss = self._distillation_loss(y_curr, y_prev, self.prev_classes)
            if self.clock.train_iterations % 10 == 0:
                print(f"COMPARING LOSSES: ce={self.loss:.4f}, kd={alpha * dist_loss:.4f}")
            return alpha * dist_loss

    def _after_training_exp(self, num_workers=0, **kwargs):
        super()._after_training_exp(**kwargs)
        # ood data
        self.ood_it = iter(DataLoader(
            self.ood_dataset,
            batch_size=128,
            num_workers=num_workers,
            shuffle=True
        ))

        # prev model
        self.prev_model = copy.deepcopy(self.model)
        self.prev_model.eval()
        self.prev_classes.extend(self.experience.classes_in_this_experience)
        assert len(self.experience.task_labels) == 1

    def next_ood_batch(self):
        try:
            mb = next(self.ood_it)
        except StopIteration as e:
            self.ood_it = iter(DataLoader(
                self.ood_dataset,
                batch_size=128,
                num_workers=8,
                shuffle=True))
            mb = next(self.ood_it)

        nmb = []
        for el in mb:
            nmb.append(el.to(self.device))
        nmb_x, _, _, _ = self.cm(nmb[0], nmb[1])  # CutMix
        nmb[0] = nmb_x
        return nmb


def main(args):
    scenario_name = "split_cifar10"
    logdir = f"$ANON_PROJECT_DIR/test/cir_distill_single_image/{scenario_name}/naive"
    os.makedirs(logdir, exist_ok=True)

    ########################
    # Scenario Preparation #
    ########################
    cir_benchmark = SplitCIFAR10(n_experiences=5, seed=1234)

    dbase = "$WORKDIRDATA/single-img-extrapolating"
    dname = "ameyoko_dataset32_singleameyoko_init0.5_deg30_scale500_1_shear30_randinterp_False_vflipFalse_cropfirstTrue_new_50000"
    STD = (0.2023, 0.1994, 0.2010)
    MEAN = (0.4914, 0.4822, 0.4465)
    transform = transforms.Compose([
        transforms.RandomCrop(32, padding=4),
        transforms.RandomHorizontalFlip(),
        transforms.ToTensor(),
        transforms.Normalize(MEAN, STD),
    ])
    train_data = DistillationDatasetFromDir(f"{dbase}/{dname}", transform)
    train_data = AvalancheClassificationDataset(train_data)

    ########################
    # Strategy Preparation #
    ########################
    student_fn = ResNet18(nclasses=10)
    # insize = student_fn.linear.weight.shape[1]
    # student_fn.linear = IncrementalClassifier(insize, masking=True)
    optimizer = torch.optim.Adam(student_fn.parameters(), lr=args.learning_rate)
    # optimizer = torch.optim.SGD(student_fn.parameters(), lr=args.learning_rate)

    strategy = OODKDStrategy(
        # KD arguments
        alpha=10,  # [0*a0, 1*a0, 2*a0, 3*a0, 4*a0, 5*a0],
        temperature=args.temperature,
        ood_dataset=train_data,
        # BaseSGD arguments
        model=student_fn,
        optimizer=optimizer,
        criterion=CrossEntropyLoss(),
        train_mb_size=args.batch_size,
        eval_mb_size=512,
        train_epochs=20,  # 100 // len(cir_benchmark.train_stream),
        device='cuda',
        eval_every=5,
    )

    ########################
    # Training Loop        #
    ########################
    NUM_WORKERS = 8
    for i, train_exp in enumerate(cir_benchmark.train_stream):
        if i > 0:  # Use real data for distillation
            ds = [exp.dataset for exp in cir_benchmark.train_stream[:i]]
            strategy.ood_dataset = AvalancheConcatClassificationDataset(ds)

        evals = [train_exp, cir_benchmark.test_stream[:i+1]]
        strategy.train(experiences=train_exp, eval_streams=evals, num_workers=NUM_WORKERS)

        eid = train_exp.current_experience
        model_fname = os.path.join(logdir, f'model_e{eid}.pth')
        torch.save(strategy.model.state_dict(), model_fname)

    strategy.eval(cir_benchmark.test_stream, num_workers=NUM_WORKERS)

    with open(os.path.join(logdir, 'metrics.json'), 'w') as f:
        f.write(to_json(strategy.evaluator.get_all_metrics(), ignore_errors=True))


if __name__ == "__main__":
    args = parser.parse_args()
    main(args)
