from functools import reduce

import setGPU

import numpy as np
import argparse
import json
import os
import re
from json import JSONDecodeError

import torch
from hydra.utils import instantiate
from omegaconf import OmegaConf
from torch.nn import CrossEntropyLoss
from torch.utils.data import DataLoader

from avalanche.benchmarks.utils import AvalancheConcatDataset
from avalanche.benchmarks.utils.dataset_utils import ConstantSequence
from avalanche.evaluation.metrics import accuracy_metrics
from avalanche.models import as_multitask
from avalanche.training import Naive
from avalanche.benchmarks import CLExperience, CLStream
from avalanche.benchmarks.utils.avalanche_dataset import _TaskSubsetDict
from avalanche.training.plugins import EvaluationPlugin
from dac.evaluation.rich_text_logging import RichTextLogger
from dac.evaluation.task_agnostic_accuracy import AvgStreamAccuracyPM
from dac.benchmarks import maclCIFAR100, SEED_BENCHMARK_RUNS, seq_cifar100_benchmark
from dac.gem_resnet import ResNet18
from avalanche.models.utils import avalanche_model_adaptation
from dac.benchmarks import as_multitask_benchmark, seq_core50_benchmark
from sklearn.linear_model import LogisticRegression

from dac.utils import get_mt_cifar100, eval_model_accuracy


def collect_latent_activations(
        model: torch.nn.Module,
        stream, latent_name='model.layer4.0'):
    model.to('cuda').eval()
    ds = [exp.dataset for exp in stream]
    data = AvalancheConcatDataset(ds)

    hs = []
    def foo(module, input, output):
        nonlocal hs
        hs.append(output.detach().cpu())
    handle = reduce(getattr, latent_name.split('.'), model).register_forward_hook(foo)

    for x, y, t in DataLoader(data, batch_size=512,
                              shuffle=False, drop_last=False):
        model(x.to('cuda'), task_labels=t.to('cuda'))
    h = torch.concat(hs, dim=0).numpy()
    handle.remove()
    model.to('cpu')
    return h.reshape(h.shape[0], -1)


def collect_data_from_stream(stream):
    ds = [exp.dataset for exp in stream]
    data = AvalancheConcatDataset(ds)
    xs, ys = [], []
    for x, y, _ in DataLoader(data, batch_size=512,
                              shuffle=False, drop_last=False):
        xs.append(x)
        ys.append(y)
    x = torch.concat(xs, dim=0).numpy()
    y = torch.concat(ys, dim=0).numpy()
    return x, y


def linear_probing(model, benchmark):
    X_train, y_train = collect_data_from_stream(benchmark.train_stream)
    H_train = collect_latent_activations(model, benchmark.train_stream)
    X_test, y_test = collect_data_from_stream(benchmark.test_stream)
    H_test = collect_latent_activations(model, benchmark.test_stream)

    probe = LogisticRegression(n_jobs=4)
    probe.fit(H_train, y_train)

    acc_train = probe.score(H_train, y_train)
    acc_test = probe.score(H_test, y_test)
    print(f"train ACC={acc_train:.4f}, test ACC={acc_test:.4f}")


def get_last_metric_value_from_json(folder, metrics_file='metrics.json',
                                    metric_name='Top1_Acc_Stream/eval_phase/test_stream/Task000'):
    """Return the desired metric value from a json file."""

    cur_file = os.path.join(folder, metrics_file)
    with open(cur_file) as f:
        # fix NaN encoding for bugged JSON config
        j = f.read()
        j = re.sub(r'\bnan\b', 'NaN', j)

        try:
            data = json.loads(j)
        except JSONDecodeError as e:
            print(f"Error decoding json file: {cur_file}")
            print(j)
            raise e

    m_eval = data[metric_name][1][-1]
    return m_eval


def average_test_stream_accuracy(folder):
    accs = []
    for i in range(5):
        logdir = os.path.join(folder, f'run{i}')
        a = get_last_metric_value_from_json(logdir)
        accs.append(a)
    accs = np.array(accs)
    return np.mean(accs), np.std(accs)


def as_multitask_stream(ss):
    new_exps = []
    for ii, e in enumerate(ss):
        ne = CLExperience()
        ne.classes_in_this_experience = e.classes_in_this_experience
        ne.task_labels = [ii]
        ne.task_label = ii
        ne.dataset = e.dataset

        ne.dataset.targets_task_labels = ConstantSequence(ii, len(ne.dataset))

        told_idx = ne.dataset.tasks_pattern_indices
        del ne.dataset.tasks_pattern_indices[0]
        ne.dataset.tasks_pattern_indices[ii] = told_idx

        task_dict = _TaskSubsetDict()
        task_idxs = ne.dataset.tasks_pattern_indices[ii]
        task_dict[ii] = (ne.dataset, task_idxs)
        ne.dataset.val_to_idx = task_dict

        new_exps.append(ne)
    return CLStream("original_test", new_exps)


def mt_eval_test_split_cifar100(folder):
    scenario = maclCIFAR100()
    scenario = as_multitask_benchmark(scenario)
    test_stream = as_multitask_stream(scenario.original_benchmark.test_stream)

    for eid in range(10):
        model = ResNet18(100)
        model = as_multitask(model, 'linear')
        for e in list(scenario.expert_models_stream)[:eid+1]:
            avalanche_model_adaptation(model, e)
        print(model)

        model.load_state_dict(torch.load(folder + f'/model_e{eid}.pth'))
        strat = Naive(model, None, criterion=CrossEntropyLoss(), eval_mb_size=512)
        strat.eval(list(test_stream))


def grid_search_mt_eval_test_split_cifar100(folder):
    scenario = maclCIFAR100()
    scenario = as_multitask_benchmark(scenario)
    test_stream = as_multitask_stream(scenario.original_benchmark.test_stream)
    model = load_resnet_mt_model(folder + f'/model_e9.pth', scenario)

    rtl = RichTextLogger('.', interactive=True)
    evalp = EvaluationPlugin(
        accuracy_metrics(experience=True),
        AvgStreamAccuracyPM(),
        loggers=[rtl])

    strat = Naive(
        model, None, criterion=CrossEntropyLoss(),
        eval_mb_size=512, evaluator=evalp)
    strat.eval(list(test_stream))


def load_resnet_mt_model(model_fname, stream):
    model = ResNet18(100)
    model = as_multitask(model, 'linear')
    for e in list(stream):
        avalanche_model_adaptation(model, e)
    # print(model)
    model.load_state_dict(torch.load(model_fname, map_location=torch.device('cpu')))
    return model.eval()


def scifar100_assessment_acc_and_std(assessdir):
    accs = []
    for runid in range(5):
        # with open(assessdir + f'/run{runid}/hydra_config.yaml') as f:
        #     cfg = OmegaConf.load(f)
        # benchmark = instantiate(cfg.benchmark.seq_benchmark_instance, run_id=cfg.run_id)
        benchmark = seq_cifar100_benchmark(num_experiences=10, run_id=runid)
        lenstream = len(benchmark.train_stream)
        fname = f"{assessdir}/run{runid}/model_e{lenstream - 1}.pth"
        model = load_resnet_mt_model(fname, benchmark.train_stream)
        tacc = eval_model_accuracy(model, benchmark.train_stream)
        acc = eval_model_accuracy(model, benchmark.test_stream)
        accs.append(acc)
        print(f"TRAIN acc={tacc:.4f}, TEST acc={acc:4f}")
    a, s = np.mean(accs), np.std(accs)
    print(f"MEAN={a:.5f}, STD={s:.5f}")
    return a, s


if __name__ == '__main__':
    base_folder = "$ANON_PROJECT_DIR"

    # exp_folder = base_folder + "/split_cifar100_mt_ood_kd/run0"
    # print(exp_folder)
    # mt_eval_test_split_cifar100(exp_folder)

    # m, s = average_test_stream_accuracy(exp_folder)
    # print(f"mean: {m}, std: {s}")

    # base_folder = "$ANON_PROJECT_DIR"
    # ms_folder = base_folder + "/split_cifar100_mt_ood_kd/ms_split_cifar100_mt_ood_kd_2022-06-29_09-54-19"
    # ms_folder = base_folder + "/split_cifar100_mt/ci_mt_oodkd/ms_split_cifar100_mt/ci_mt_oodkd_2022-07-4_11-55-54"
    # print(ms_folder)
    # for dd in os.listdir(ms_folder):
    #     print(dd)
    #     if os.path.isdir(f"{ms_folder}/{dd}"):
    #         print(dd)
    #         if not os.path.exists(f"{ms_folder}/{dd}/run0" + f'/model_e9.pth'):
    #             continue
    #         # grid_search_mt_eval_test_split_cifar100(f"{ms_folder}/{dd}/run0")
    #
    #         scenario = maclCIFAR100()
    #         scenario = as_multitask_benchmark(scenario)
    #         test_stream = as_multitask_stream(scenario.original_benchmark.test_stream)
    #         model = load_mt_model(f"{ms_folder}/{dd}/run0" + f'/model_e9.pth', scenario)
    #         linear_probing(model, scenario)

    # ms_folder = "/data2/users/auname/dac_project" + "/split_cifar100_mt/ci_mt_oodkd_2022-07-13"
    # print(ms_folder)
    # for eid in range(10):
    #     scenario = get_mt_cifar100(seed=SEED_BENCHMARK_RUNS[0])
    #     model = load_mt_model(f"{ms_folder}/run0" + f'/model_e{eid}.pth', scenario.train_stream)
    #     eval_model_accuracy(model, scenario.test_stream)
    #     # linear_probing(model, scenario)

    #################################################
    # EVAL ASSESSMENT
    #################################################
    # assessdir = base_folder + "/split_cifar100_mt/cl_mt_latproj--best/assess_2022-08-5_11-44-01"  # MEAN=0.78216, STD=0.00699
    # assessdir = base_folder + "/cifar100_T5/cl_mt_latproj/assess_2022-08-18_10-59-58"  # MEAN=0.72610, STD=0.01023

    ######## CIFAR100 - T20 ########
    # assessdir = base_folder + "/cifar100_T20/cl_mt_latproj/assess_2022-09-6_14-50-10"  # MEAN=0.86236, STD=0.01050

    ######## CIFAR100 - T5 ########
    # assessdir = base_folder + "/cifar100_T5/cl_mt_latproj/assess_2022-09-6_14-50-10"  # RUNNING

    ######## ABLATION ########
    # assessdir = base_folder + "/split_cifar100/seq_kd_curr/assess_2022-09-5_13-07-54"  # MEAN=0.84252, STD=0.00585
    # assessdir = base_folder + "/ablation/ood_data/ameyoko/assess_2022-09-2_14-47-15"  # MEAN=0.80580, STD=0.00857
    # assessdir = base_folder + "/ablation/ood_data/animals/assess_2022-09-15_09-16-55"  # MEAN=0.82102, STD=0.00595
    # assessdir = base_folder + "/ablation/ood_data/bridge/assess_2022-09-13_15-55-58"  # MEAN=0.79046, STD=0.00642
    # assessdir = base_folder + "/ablation/ood_data/hubble/assess_2022-09-09_17-59-35"  # MEAN=0.23315, STD=0.21852 very weird results. One exp got 60%, the others 10%
    # assessdir = base_folder + "/ablation/ood_data/imagenet/assess_2022-09-2_14-47-15"
    # assessdir = base_folder + "/ablation/ood_data/noise/assess_2022-09-15"  # MEAN=0.10774, STD=0.00503

    scifar100_assessment_acc_and_std(assessdir)
