import os

import setGPU
import matplotlib.pyplot as plt
import tueplots.bundles

from dac.benchmarks import maclCIFAR100
from experiments.eval_forward_transfer import load_experts, maybe_compute, load_independent_experts
from torch_cka import CKA
from torch.utils.data import DataLoader
import numpy as np
import random

from dac.utils import plotstyle_iclr2022


def pairwise_cka(ms1, ms2, data):
    def seed_worker(worker_id):
        np.random.seed(0)
        random.seed(0)

    dataloader = DataLoader(data, batch_size=128, shuffle=False, worker_init_fn=seed_worker)
    res = []
    for m1, m2 in zip(ms1, ms2):
        m1.to('cuda'), m2.to('cuda')
        cka = CKA(m1, m2, device='cuda')
        cka.compare(dataloader)
        res.append(cka.export()['CKA'])
        m1.to('cpu'), m2.to('cpu')
    return res


def plot_ckas(ckas):
    fig, axs = plt.subplots(1, len(ckas), figsize=(22, 5), sharey=True, sharex=True)
    for i, r in enumerate(ckas):
        axs[i].matshow(r, vmin=0, vmax=1.0)
        axs[i].set_xticks([])
        axs[i].set_yticks([])


if __name__ == '__main__':
    basedir = "$ANON_PROJECT_DIR"
    logdir = '$ANON_PROJECT_DIR/plots/cka'
    os.makedirs(logdir, exist_ok=True)
    os.chdir(logdir)

    # Using seaborn's style
    # plt.style.use('seaborn')
    print(tueplots.bundles.neurips2022())
    plt.rcParams.update(tueplots.bundles.neurips2022())

    macl_scenario = maclCIFAR100()
    macl_models = [e.expert_model for e in macl_scenario.expert_models_stream]
    dataT9 = macl_scenario.original_benchmark.test_stream[9].dataset

    ## SameInit
    ms1 = [macl_models[-1] for _ in range(9)]
    ckas_sameinit = maybe_compute(pairwise_cka, './ckas_same.pkl', ms1=ms1, ms2=macl_models[:-1], data=dataT9)
    plot_ckas(ckas_sameinit)
    plt.savefig("./ckas_same.png")
    plt.savefig("./ckas_same.pdf")

    ## RandInit
    rundir = basedir + "/pretrained_experts/split_cifar100_random_init/run0"
    randnit_models = load_independent_experts(rundir)
    ms1 = [randnit_models[-1] for _ in range(9)]
    ckas_randinit = maybe_compute(pairwise_cka, './ckas_rand.pkl', ms1=ms1, ms2=randnit_models[:-1], data=dataT9)
    plot_ckas(ckas_randinit)
    plt.savefig("./ckas_rand.png")
    plt.savefig("./ckas_rand.pdf")

    ## Seq-OODKD
    rundir = basedir + "/split_cifar100_mt/ci_mt_oodkd -- optimal_expert_hparams/" \
        "ms_split_cifar100_mt/ci_mt_oodkd -- optimal_expert_hparams_2022-07-24_19-35-29" \
        "/override_hps_and_train_05af1_00003_3_optimizer.lr=0.001,strategy.init_from_first=False,strategy.temperature=0.5_2022-07-24_19-35-37/run0"

    seq_oodkd_models = load_experts(rundir)
    ms1 = [seq_oodkd_models[-1] for _ in range(9)]
    ckas_seq_oodkd = maybe_compute(pairwise_cka, './ckas_seq_oodkd.pkl', ms1=ms1, ms2=seq_oodkd_models[:-1], data=dataT9)
    plot_ckas(ckas_seq_oodkd)
    plt.savefig("./ckas_seq_oodkd.png", bbox_inches='tight')
    plt.savefig("./ckas_seq_oodkd.pdf", bbox_inches='tight')

    ## Seq-LatProj
    expdir = basedir + "/split_cifar100_mt/cl_mt_latproj--project_l4_and_linear/ms_2022-08-2_10-31-33"
    rundir = expdir + "/override_hps_and_train_86dfe_00000_0_optimizer.lr=0.0001,strategy.latent_proj_lambda=100.0_2022-08-02_10-31-38"

    seq_latproj_models = load_experts(rundir)
    ms1 = [seq_latproj_models[-1] for _ in range(9)]
    ckas_seq_latproj = maybe_compute(pairwise_cka, './ckas_seq_latproj.pkl', ms1=ms1, ms2=seq_latproj_models[:-1], data=dataT9)
    plot_ckas(ckas_seq_latproj)
    plt.savefig("./ckas_seq_latproj.png", bbox_inches='tight')
    plt.savefig("./ckas_seq_latproj.pdf", bbox_inches='tight')

    fig, axs = plt.subplots(1, 4, figsize=(4.0, 16.0), sharey=True, sharex=True)

    idx = 0
    # axs[idx].matshow(ckas_sameinit[0].export()['CKA'], vmin=0, vmax=1.0)
    axs[idx].matshow(ckas_sameinit[0].numpy(), vmin=0, vmax=1.0)
    axs[idx].set_title("Independent\nCommon Init", multialignment='center')
    axs[idx].set_xticks([])
    axs[idx].set_yticks([])

    idx = 1
    # axs[idx].matshow(ckas_randinit[0].export()['CKA'], vmin=0, vmax=1.0)
    axs[idx].matshow(ckas_randinit[0].numpy(), vmin=0, vmax=1.0)
    axs[idx].set_title("Independent\nRandom Init", multialignment='center')
    axs[idx].set_xticks([])
    axs[idx].set_yticks([])

    idx = 2
    # axs[idx].matshow(ckas_seq_oodkd[0].export()['CKA'], vmin=0, vmax=1.0)
    axs[idx].matshow(ckas_seq_oodkd[0].numpy(), vmin=0, vmax=1.0)
    axs[idx].set_title("Sequential\nw.o. PLD", multialignment='center')
    axs[idx].set_xticks([])
    axs[idx].set_yticks([])

    idx = 3
    # axs[idx].matshow(ckas_seq_latproj[0].export()['CKA'], vmin=0, vmax=1.0)
    axs[idx].matshow(ckas_seq_latproj[0].numpy(), vmin=0, vmax=1.0)
    axs[idx].set_title("Sequential\nDFC", multialignment='center')
    axs[idx].set_xticks([])
    axs[idx].set_yticks([])

    plt.savefig('./cka_combined.png', bbox_inches='tight')
    plt.savefig('./cka_combined.pdf', bbox_inches='tight')
