"""
    Forward transfer plot for MACL v2.
"""
import os.path
import pickle

import setGPU

import matplotlib.pyplot as plt
from dac.benchmarks import maclCIFAR100
from avalanche.models import SlimResNet18
import torch
from avalanche.training import Naive
from avalanche.training.plugins import EvaluationPlugin
from dac.evaluation.task_agnostic_accuracy import AvgStreamAccuracyPM
import numpy as np
from experiments.eval_linear_probing import collect_numpy_data_from_stream, collect_latent_activations_from_stream
from sklearn.linear_model import LogisticRegression, SGDClassifier
from sklearn.neighbors import KNeighborsClassifier
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import StandardScaler
from tueplots import bundles


def maybe_compute(foo, fname, **args):
    if os.path.exists(fname):
        with open(fname, 'rb') as f:
            return pickle.load(f)
    else:
        res = foo(**args)
        with open(fname, 'wb') as f:
            pickle.dump(res, f)
        return res


def getacc(model, stream):
    evalp = EvaluationPlugin(AvgStreamAccuracyPM(), loggers=[])
    strategy = Naive(model, None, eval_mb_size=512, device='cuda', evaluator=evalp)
    strategy.eval(stream, num_workers=8)
    mname = f"AvgStreamAcc/eval_phase/{stream[0].origin_stream.name}_stream"
    streamacc = strategy.evaluator.get_all_metrics()[mname][1][-1]
    return streamacc


def load_experts(rundir):
    seq_models = []
    for eid in range(10):
        mname = rundir + f"/expert_e{eid}.pth"
        model = SlimResNet18(nclasses=100)
        model.load_state_dict(torch.load(mname, map_location=torch.device('cpu')))
        seq_models.append(model)
    return seq_models


def get_train_test_accs(models, train_stream, test_stream):
    train_accs = [getacc(model, [exp]) for model, exp in zip(models, train_stream)]
    print("TRAIN ACCS: ", train_accs)
    test_accs = [getacc(model, [exp]) for model, exp in zip(models, test_stream)]
    print("TEST ACCS: ", test_accs)
    return train_accs, test_accs


def linear_probing(model, train_stream, test_stream, probe_type="linear", layer_name="linear"):
    X_train, y_train = collect_numpy_data_from_stream(list(train_stream)[:])
    H_train = collect_latent_activations_from_stream(model, list(train_stream)[:], latent_name=layer_name)
    X_test, y_test = collect_numpy_data_from_stream(list(test_stream)[:])
    H_test = collect_latent_activations_from_stream(model, list(test_stream)[:], latent_name=layer_name)

    scaler = StandardScaler()
    if probe_type == "linear":
        probe = LogisticRegression(multi_class='multinomial', n_jobs=-1)
    elif probe_type == "knn":
        probe = KNeighborsClassifier(n_neighbors=40)
    else:
        assert False
    model = make_pipeline(scaler, probe)
    model.fit(H_train, y_train)

    acc_train = model.score(H_train, y_train)
    acc_test = model.score(H_test, y_test)
    print(f"train ACC={acc_train:.4f}, test ACC={acc_test:.4f}")
    return acc_train, acc_test


def load_independent_experts(rundir, len_stream=10):
    experts_stream = []
    for i in range(len_stream):
        fname_i = f"{rundir}/model_e{i}.pt"
        experts_stream.append(torch.load(fname_i).to("cpu").eval())
    return experts_stream


if __name__ == '__main__':
    logdir = '$ANON_PROJECT_DIR/plots/forward_transfer'
    basedir = "$ANON_PROJECT_DIR"
    os.makedirs(logdir, exist_ok=True)
    os.chdir(logdir)

    macl_scenario = maclCIFAR100()
    train_stream = macl_scenario.original_benchmark.train_stream
    test_stream = macl_scenario.original_benchmark.test_stream

    ###########################################
    # Common INIT macl models
    ###########################################
    print("SAME INIT:")
    sameinit_models = [e.expert_model for e in macl_scenario.expert_models_stream]
    sameinit_train_accs, sameinit_test_accs = maybe_compute(
        get_train_test_accs,
        './sameinit_models_accs.pkl',
        models=sameinit_models,
        train_stream=train_stream,
        test_stream=test_stream)

    ###########################################
    # Random INIT macl models
    ###########################################
    print("RANDOM INIT:")
    rundir = basedir + "/pretrained_experts/split_cifar100_random_init/run0"
    randnit_models = load_independent_experts(rundir)
    randinit_train_accs, randinit_test_accs = maybe_compute(
        get_train_test_accs,
        './randinit_models_accs.pkl',
        models=randnit_models,
        train_stream=train_stream,
        test_stream=test_stream)

    ###########################################
    # Sequential-macl - LatProj
    ###########################################
    expdir = basedir + "/split_cifar100_mt/cl_mt_latproj--project_l4_and_linear/ms_2022-08-2_10-31-33"
    rundir = expdir + "/override_hps_and_train_86dfe_00000_0_optimizer.lr=0.0001,strategy.latent_proj_lambda=100.0_2022-08-02_10-31-38"

    print("Sequential LatProj")
    seq_latproj_models = load_experts(rundir)
    seq_latproj_train_accs, seq_latproj_test_accs = maybe_compute(
        get_train_test_accs, './seq_latproj_accs.pkl',
        models=seq_latproj_models, train_stream=train_stream, test_stream=test_stream)

    ###########################################
    # Sequential-macl - OODKD
    ###########################################
    basedir = "$ANON_PROJECT_DIR"
    rundir = basedir + "/split_cifar100_mt/ci_mt_oodkd -- optimal_expert_hparams/" \
        "ms_split_cifar100_mt/ci_mt_oodkd -- optimal_expert_hparams_2022-07-24_19-35-29" \
        "/override_hps_and_train_05af1_00003_3_optimizer.lr=0.001,strategy.init_from_first=False,strategy.temperature=0.5_2022-07-24_19-35-37/run0"

    print("Sequential OODKD")
    seq_oodkd_models = load_experts(rundir)
    seq_oodkd_train_accs, seq_oodkd_test_accs = maybe_compute(
        get_train_test_accs, './seq_oodkd_models.pkl',
        models=seq_oodkd_models, train_stream=train_stream, test_stream=test_stream)

    print(bundles.neurips2022())
    plt.rcParams.update(bundles.neurips2022())

    plt.figure(figsize=(3.07, 1.91))
    w = 0.2
    idxs = np.array(list(range(11)))
    plt.bar(idxs - 1.5*w, sameinit_test_accs + [np.mean(sameinit_test_accs)], w, label="same-init")
    plt.bar(idxs - 0.5*w, randinit_test_accs + [np.mean(randinit_test_accs)], w, label="random-init")
    plt.bar(idxs + 0.5*w, seq_oodkd_test_accs + [np.mean(seq_oodkd_test_accs)], w, label="seq-OODKD")
    plt.bar(idxs + 1.5*w, seq_latproj_test_accs + [np.mean(seq_latproj_test_accs)], w, label="seq-PLD")
    plt.xticks(list(range(11)), [f"E{eid}" for eid in range(10)] + ["avg"])
    plt.ylabel("Accuracy")
    plt.xlabel("Expert")
    plt.legend(loc="lower right")
    plt.savefig('./pbar_curr_task.png')
    plt.savefig('./pbar_curr_task.pdf')

    #################################################
    # PROBING ON ALL THE DATA
    #################################################
    def probe_all(models):
        accs = []
        for idx, mod in enumerate(models):
            print(f"{idx} ", end='')
            accs.append(linear_probing(mod, train_stream, test_stream)[1])
        return accs

    sameinit_accs_all = maybe_compute(probe_all, './sameinit_all_accs', models=sameinit_models)
    randinit_accs_all = maybe_compute(probe_all, './randinit_all_accs', models=randnit_models)
    seq_oodkd_accs_all = maybe_compute(probe_all, './seq_oodkd_all_accs', models=seq_oodkd_models)
    seq_latproj_accs_all = maybe_compute(probe_all, './seq_latproj_all_accs', models=seq_latproj_models)
    print("sameinit: ", sameinit_accs_all)
    print("randinit: ", randinit_accs_all)
    print("SEQ-OODKD: ", seq_oodkd_accs_all)
    print("SEQ-PLD: ", seq_latproj_accs_all)

    # plt.rcParams.update(bundles.neurips2022())
    plt.figure(figsize=(3.07, 1.91))
    w = 0.2
    idxs = np.array(list(range(11)))
    plt.bar(idxs - 1.5*w, sameinit_accs_all + [np.mean(sameinit_accs_all)], w, label="Independent-common-init")
    plt.bar(idxs - 0.5*w, randinit_accs_all + [np.mean(randinit_accs_all)], w, label="Independent-random-init")
    plt.bar(idxs + 0.5*w, seq_oodkd_accs_all + [np.mean(seq_oodkd_accs_all)], w, label="Seq-w.o. PLD")
    plt.bar(idxs + 1.5*w, seq_latproj_accs_all + [np.mean(seq_latproj_accs_all)], w, label="Seq-PLD")
    plt.xticks(list(range(11)), [f"E{eid}" for eid in range(10)] + ["avg"])
    plt.ylabel("Accuracy")
    plt.xlabel("Expert")
    plt.legend(loc="lower right")
    plt.savefig('./pbar_all_task.png')
    plt.savefig('./pbar_all_task.pdf')
