import copy
from functools import reduce
from types import SimpleNamespace

import setGPU

import torch
from sklearn.neighbors import KNeighborsClassifier
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import StandardScaler
from sklearn.svm import LinearSVC
from torch.utils.data import DataLoader
from tqdm import tqdm

from avalanche.benchmarks.utils import AvalancheConcatDataset
from avalanche.benchmarks.utils.dataset_utils import ConstantSequence
from avalanche.models import as_multitask
from avalanche.benchmarks import CLExperience, CLStream
from avalanche.benchmarks.utils.avalanche_dataset import _TaskSubsetDict, AvalancheSubset, AvalancheDataset
from dac.benchmarks import SEED_BENCHMARK_RUNS, maclCIFAR100
from dac.gem_resnet import ResNet18
from avalanche.models.utils import avalanche_model_adaptation, avalanche_forward
from sklearn.linear_model import LogisticRegression, SGDClassifier

from dac.utils import get_mt_cifar100, eval_model_accuracy
from dac.training.synthetic.utils import IncrementalImageDataset
from avalanche.benchmarks.classic.ccifar10 import _default_cifar10_train_transform
import numpy as np


def collect_latent_activations_from_stream(
        model: torch.nn.Module,
        stream, latent_name='linear'):
    model.to('cuda').eval()
    ds = [exp.dataset for exp in stream]
    data = AvalancheConcatDataset(ds)
    return collect_latent_activations_from_data(data, latent_name, model)


def collect_latent_activations_from_data(data, latent_name, model):
    model = model.eval().cuda()
    hs = []

    def foo(module, input, output):
        nonlocal hs
        hs.append(input[0].detach().cpu())

    handle = reduce(getattr, latent_name.split('.'), model).register_forward_hook(foo)
    for x, y, t in DataLoader(data, batch_size=512,
                              shuffle=False, drop_last=False):
        avalanche_forward(model, x.to('cuda'), t.to('cuda'))
    h = torch.concat(hs, dim=0).numpy()
    handle.remove()
    model.to('cpu')
    return h.reshape(h.shape[0], -1)


def collect_numpy_data_from_stream(stream):
    ds = [exp.dataset for exp in stream]
    data = AvalancheConcatDataset(ds)
    return collect_numpy_data(data)


def collect_numpy_data(data):
    xs, ys = [], []
    for x, y, _ in DataLoader(data, batch_size=512,
                              shuffle=False, drop_last=False):
        xs.append(x)
        ys.append(y)
    x = torch.concat(xs, dim=0).numpy()
    y = torch.concat(ys, dim=0).numpy()
    return x, y


def linear_probing(model, benchmark, probe_type="linear", layer_name="linear"):
    X_train, y_train = collect_numpy_data_from_stream(list(benchmark.train_stream)[:])
    H_train = collect_latent_activations_from_stream(model, list(benchmark.train_stream)[:], latent_name=layer_name)
    X_test, y_test = collect_numpy_data_from_stream(list(benchmark.test_stream)[:])
    H_test = collect_latent_activations_from_stream(model, list(benchmark.test_stream)[:], latent_name=layer_name)

    scaler = StandardScaler()
    if probe_type == "linear":
        probe = LogisticRegression(multi_class='multinomial', n_jobs=-1)
    elif probe_type == "knn":
        probe = KNeighborsClassifier(n_neighbors=40)
    else:
        assert False
    model = make_pipeline(scaler, probe)
    model.fit(H_train, y_train)

    acc_train = model.score(H_train, y_train)
    acc_test = model.score(H_test, y_test)
    print(f"train ACC={acc_train:.4f}, test ACC={acc_test:.4f}")
    return acc_train, acc_test


def linear_probing_multitask(model, benchmark, probe_type="linear"):
    train_accs, test_accs = [], []
    for eid, exp in enumerate(benchmark.train_stream):
        X_train, y_train = collect_numpy_data(exp.dataset)
        H_train = collect_latent_activations_from_data(exp.dataset, 'linear', model)
        X_test, y_test = collect_numpy_data(benchmark.test_stream[eid].dataset)
        H_test = collect_latent_activations_from_data(benchmark.test_stream[eid].dataset, 'linear', model)

        scaler = StandardScaler()
        if probe_type == "linear":
            probe = LogisticRegression(multi_class='multinomial', C=0.01, n_jobs=-1)
        elif probe_type == "knn":
            probe = KNeighborsClassifier()
        else:
            assert False
        skmodel = make_pipeline(scaler, probe)
        skmodel.fit(H_train, y_train)

        train_accs.append(skmodel.score(H_train, y_train))
        test_accs.append(skmodel.score(H_test, y_test))
    print(f"train ACC={np.mean(train_accs):.4f}, test ACC={np.mean(test_accs):.4f}")


def knn_probing_multitask(model, benchmark):
    train_accs, test_accs = [], []
    for eid, exp in enumerate(benchmark.train_stream):
        X_train, y_train = collect_numpy_data(exp.dataset)
        H_train = collect_latent_activations_from_data(exp.dataset, 'linear', model)
        X_test, y_test = collect_numpy_data(benchmark.test_stream[eid].dataset)
        H_test = collect_latent_activations_from_data(benchmark.test_stream[eid].dataset, 'linear', model)

        scaler = StandardScaler()
        knn = KNeighborsClassifier()
        skmodel = make_pipeline(scaler, knn)
        skmodel.fit(H_train, y_train)

        train_accs.append(skmodel.score(H_train, y_train))
        test_accs.append(skmodel.score(H_test, y_test))
    print(f"train ACC={np.mean(train_accs):.4f}, test ACC={np.mean(test_accs):.4f}")


def as_multitask_stream(ss):
    new_exps = []
    for ii, e in enumerate(ss):
        ne = CLExperience()
        ne.classes_in_this_experience = e.classes_in_this_experience
        ne.task_labels = [ii]
        ne.task_label = ii
        ne.dataset = e.dataset

        ne.dataset.targets_task_labels = ConstantSequence(ii, len(ne.dataset))

        told_idx = ne.dataset.tasks_pattern_indices
        del ne.dataset.tasks_pattern_indices[0]
        ne.dataset.tasks_pattern_indices[ii] = told_idx

        task_dict = _TaskSubsetDict()
        task_idxs = ne.dataset.tasks_pattern_indices[ii]
        task_dict[ii] = (ne.dataset, task_idxs)
        ne.dataset.val_to_idx = task_dict

        new_exps.append(ne)
    return CLStream("original_test", new_exps)


def load_mt_model(model_fname, stream):
    model = ResNet18(100)
    model = as_multitask(model, 'linear')
    for e in list(stream):
        avalanche_model_adaptation(model, e)
    # print(model)
    model.load_state_dict(torch.load(model_fname))
    return model


class LabeledData:
    def __init__(self, data, targets):
        self.data = data
        self.targets = targets

    def __getitem__(self, idx):
        return self.data[idx], self.targets[idx]

    def __len__(self):
        return len(self.targets)


def make_synthetic_data_stream(scenario, dataroot):
    synthetic_datas = []
    for eid, exp in enumerate(scenario.expert_models_stream):
        syntfname = dataroot + f"/img_pool_e{eid}"
        sd = IncrementalImageDataset(syntfname, transform=_default_cifar10_train_transform)
        sd = AvalancheDataset(sd)

        esd = SimpleNamespace(
            dataset=sd,
            classes_in_this_experience=np.unique(sd.targets).tolist(),
            task_labels=[eid],
            task_label=eid
        )
        synthetic_datas.append(esd)
    stream = CLStream(name="train", exps_iter=synthetic_datas)
    return stream


def make_synthetic_data_stream_single_model(dataroot):
    sd = IncrementalImageDataset(dataroot + f"/img_pool", transform=_default_cifar10_train_transform)
    esd = SimpleNamespace(
        dataset=AvalancheDataset(sd),
        classes_in_this_experience=np.unique(sd.targets).tolist(),
        task_labels=[0],
        task_label=0
    )
    stream = CLStream(name="train", exps_iter=[esd])
    return stream


if __name__ == '__main__':
    # basedir = "/data2/users/auname/dac_project"
    basedir = "$ANON_PROJECT_DIR"
    # datadir = basedir + "/synthesis/split_cifar100/deep_inversion"
    datadir = basedir + "/synthesis/split_cifar100/deep_inversion_latproj_final"

    # # Best for OODKD
    # ms_folder = basedir + "/split_cifar100_mt/ci_mt_oodkd -- optimal_expert_hparams"\
    #                       "/ms_split_cifar100_mt/ci_mt_oodkd -- optimal_expert_hparams_2022-07-24_19-35-29"
    # ms_folder += "/override_hps_and_train_05af1_00003_3_optimizer.lr=0.001,strategy.init_from_first=False,strategy.temperature=0.5_2022-07-24_19-35-37"

    # Best for OODKD-LatProj
    ms_folder = basedir + "/split_cifar100_mt/cl_mt_latproj--init_first/ms_split_cifar100_mt/cl_mt_latproj--init_first_2022-07-25_10-33-01" \
                          "/override_hps_and_train_68287_00000_0_optimizer.lr=0.0001,strategy.latent_proj_lambda=0.001_2022-07-25_10-33-06"

    cl_scenario = get_mt_cifar100(seed=SEED_BENCHMARK_RUNS[0])
    macl_scenario = maclCIFAR100(scenario="split")

    syntethic_scenario = get_mt_cifar100(seed=SEED_BENCHMARK_RUNS[0])
    # synthetic_stream = make_synthetic_data_stream(macl_scenario, datadir)
    synthetic_stream = make_synthetic_data_stream_single_model(datadir)
    syntethic_scenario.train_stream = synthetic_stream

    print(ms_folder)
    for eid in range(10):
        if eid != 9:
            continue

        model = load_mt_model(f"{ms_folder}/run0" + f'/model_e{eid}.pth', cl_scenario.train_stream)
        # model = load_mt_model(f"{ms_folder}/run0" + f'/expert_e{eid}.pth', cl_scenario.train_stream)

        # acc = eval_model_accuracy(model, cl_scenario.test_stream)
        # print(f"(EID={eid}) Test ACC={acc:.4f}")
        # linear_probing(model, cl_scenario, probe_type="linear", layer_name="model.layer4.0")
        # linear_probing_multitask(model, cl_scenario, probe_type="linear")

        ## Probes trained on SYNTHETIC data
        linear_probing(model, syntethic_scenario, probe_type="knn", layer_name="linear")
        # linear_probing_multitask(model, syntethic_scenario, probe_type="linear")
