import copy
import os

import torch.optim
from torch.nn import CrossEntropyLoss
from torchvision import transforms
from avalanche.benchmarks.utils import AvalancheClassificationDataset
from avalanche.benchmarks import SplitCIFAR10
from avalanche.evaluation.metric_results import MetricValue
from avalanche.evaluation.metrics import accuracy_metrics, loss_metrics
from avalanche.logging import InteractiveLogger, TensorboardLogger
from avalanche.models import MultiTaskModule
from avalanche.training import Naive
from avalanche.training.plugins import EvaluationPlugin
from dac.gem_resnet import ResNet18
from experiments.distill_single_image import OODKDStrategy

from dac.ood_data import DistillationDatasetFromDir
from dac.utils import to_json
import datetime
from torch import nn
from torch.utils.data import DataLoader
import torch.nn.functional as F


def masked_kd_loss(student_logits, teacher_logits, active_units, temperature):
    """
    Compute distillation loss between output of the current model and
    and output of the previous (saved) model.
    """
    curr_out = student_logits[:, active_units] / temperature
    prev_out = teacher_logits[:, active_units] / temperature
    res = F.kl_div(
        F.log_softmax(curr_out, dim=1),
        F.log_softmax(prev_out, dim=1),
        log_target=True,
        reduction="batchmean")
    return res


def masked_ce_loss(y_pred, y_true, active_units, mask_value=-10):
    mask = torch.ones(y_pred.shape[-1])
    mask[active_units] = 0
    mask = mask.bool()
    # print(mask.tolist())
    y_pred[:, mask] = mask_value
    return F.cross_entropy(y_pred, y_true)


class GeneralizedKDStrategy(Naive):
    def __init__(self, alpha, temperature,
                 ce_units, kd_units, **kwargs):
        """Generalized KD strategy.

        `kd_dataset` can be updated by outside code or plugins (during after_training_exp).

        No CutMix here.
        No OOD data here.
        Just a parametric choice for the active units to penalize.

        :param alpha:
        :param temperature:
        :param kd_dataset:
        :param kwargs:
        """
        super().__init__(**kwargs)
        self.alpha = alpha
        self.temperature = temperature

        assert ce_units in {'all', 'new'}
        assert kd_units in {'all', 'old'}
        self.ce_units = ce_units
        self.kd_units = kd_units

        # Strategy state
        self.prev_model = None
        self.prev_classes = []
        self.all_classes = []

    def criterion(self):
        if not self.is_training:
            return F.cross_entropy(self.mb_output, self.mb_y)
        else:  # training
            ce_loss = masked_ce_loss(self.mb_output, self.mb_y, self.get_units(self.ce_units))
            # KD loss computation here
            alpha = (
                self.alpha[self.clock.train_exp_counter]
                if isinstance(self.alpha, (list, tuple))
                else self.alpha
            )
            kd_student_pred = self.mb_output

            if self.prev_model is None:
                kd_loss = torch.tensor(0)
            else:
                with torch.no_grad():
                    assert not isinstance(self.prev_model, MultiTaskModule)
                    kd_teacher_pred = self.prev_model(self.mb_x)
                kd_loss = masked_kd_loss(kd_student_pred, kd_teacher_pred,
                                         self.get_units(self.kd_units), self.temperature)

            mval = MetricValue(self, 'ce', ce_loss.detach().cpu().item(), self.clock.train_iterations)
            self.evaluator.publish_metric_value(mval)
            mval = MetricValue(self, 'kd', kd_loss.detach().cpu().item(), self.clock.train_iterations)
            self.evaluator.publish_metric_value(mval)
            return ce_loss + alpha * kd_loss

    def _after_training_exp(self, **kwargs):
        super()._after_training_exp(**kwargs)
        # prev model
        self.prev_model = copy.deepcopy(self.model)
        self.prev_model.eval()
        self.prev_classes.extend(self.experience.classes_in_this_experience)
        assert len(self.experience.task_labels) == 1

    def _before_training_exp(self, **kwargs):
        super()._before_training_exp(**kwargs)
        self.all_classes = self.prev_classes + self.experience.classes_in_this_experience

    def get_units(self, config):
        if config == 'all':
            return self.all_classes
        elif config == 'old':
            return self.prev_classes
        elif config == 'new':
            return self.experience.classes_in_this_experience
        else:
            assert False, "Invalid units masking config."


def main():
    scenario_name = "split_cifar10"
    base_dir = f"$ANON_PROJECT_DIR/kd_units_choice/{scenario_name}/generalized_kd"

    for ce_units in ["all", "new"]:
        for kd_units in ["all", "old"]:
            for alpha in [0.1, 1.0, 2.0, 4.0]:
                sdate = datetime.datetime.now().strftime("%Y-%m-%-d_%H-%M-%S")
                rundir = f"run{sdate}_ce{ce_units}_kd{kd_units}_alpha{alpha}"
                logdir = f"{base_dir}/{rundir}"
                os.makedirs(logdir, exist_ok=True)

                ########################
                # Scenario Preparation #
                ########################
                cir_benchmark = SplitCIFAR10(n_experiences=5, seed=1234)

                ########################
                # Strategy Preparation #
                ########################
                student_fn = ResNet18(nclasses=10)
                optimizer = torch.optim.Adam(student_fn.parameters(), lr=0.001)
                # optimizer = torch.optim.SGD(student_fn.parameters(), lr=0.001)

                evaluator = EvaluationPlugin(
                    accuracy_metrics(minibatch=False, epoch=True, experience=True, stream=True),
                    loss_metrics(minibatch=False, epoch=True, experience=True, stream=True),
                    loggers=[InteractiveLogger(), TensorboardLogger(logdir)],
                    suppress_warnings=True,
                )

                strategy = GeneralizedKDStrategy(
                    # KD arguments
                    alpha=alpha,
                    temperature=2.0,
                    ce_units=ce_units,
                    kd_units=kd_units,
                    # BaseSGD arguments
                    model=student_fn,
                    optimizer=optimizer,
                    criterion=CrossEntropyLoss(),
                    train_mb_size=64,
                    eval_mb_size=512,
                    train_epochs=100,  # 100 // len(cir_benchmark.train_stream),
                    device='cuda',
                    eval_every=5,
                    evaluator=evaluator
                )

                ########################
                # Training Loop        #
                ########################
                NUM_WORKERS = 8
                for i, train_exp in enumerate(cir_benchmark.train_stream):
                    evals = [train_exp, cir_benchmark.test_stream[:i+1]]
                    strategy.train(experiences=train_exp, eval_streams=evals, num_workers=NUM_WORKERS)

                    eid = train_exp.current_experience
                    model_fname = os.path.join(logdir, f'model_e{eid}.pth')
                    torch.save(strategy.model.state_dict(), model_fname)
                strategy.eval(cir_benchmark.test_stream, num_workers=NUM_WORKERS)

                with open(os.path.join(logdir, 'metrics.json'), 'w') as f:
                    f.write(to_json(strategy.evaluator.get_all_metrics(), ignore_errors=True))


if __name__ == "__main__":
    main()
